from five import grok

from Products.CMFPlone.PloneBatch import Batch

from seantis.plonetools import tools

from seantis.people import _
from seantis.people.interfaces import IList
from seantis.people.browser import BaseView, Renderer
from seantis.people.content.list import ListFilter, LetterFilter
from seantis.people.supermodel import (
    get_schema_columns, get_compound_columns
)


class ReindexListView(BaseView):
    """ View to work around issue #10:
    https://github.com/seantis/seantis.people/issues/10

    """

    permission = 'cmf.ManagePortal'
    grok.require(permission)
    grok.context(IList)
    grok.name('reindex')

    def render(self):
        self.context.reindex_organisations()
        return 'objects reindexed'


class ListView(BaseView):

    permission = 'zope2.View'
    grok.require(permission)
    grok.context(IList)
    grok.name('view')

    template = grok.PageTemplateFile('templates/list.pt')
    filter_prefix = 'filter-'

    def update(self):
        if self.has_people:
            self.renderer = Renderer(
                self.schema, redirects=get_compound_columns(), place='list'
            )

    @property
    def has_people(self):
        return bool(self.context.used_type())

    @property
    def schema(self):
        assert self.has_people, """
            Do not use if there are no people in the list.
        """

        return self.context.used_type().lookupSchema()

    @property
    def filter(self):
        if self.letter:
            return LetterFilter(self.letter, _(u'Letter'))

        for key in self.request.keys():
            if key.startswith(self.filter_prefix):
                val = self.request[key].decode('utf-8')
                key = key.replace(self.filter_prefix, '')

                field = self.has_people and self.schema.get(key, None) or None
                title = field and field.title or u''

                return ListFilter(key, val, title)

        return None

    @property
    def letter(self):
        return self.request.get('letter', '').decode('utf-8')

    @property
    def batch_size(self):
        return int(self.request.get('b_size', self.context.people_per_page))

    @property
    def batch_start(self):
        return int(self.request.get('b_start', 0))

    def people(self):
        people = self.context.people(
            filter=self.filter,
            batch_start=self.batch_start,
            batch_size=self.batch_size
        )
        return Batch(people, self.batch_size, self.batch_start)

    def columns(self):
        return get_schema_columns(self.schema, self.context)

    def column_values(self, column):
        assert column.selectable

        people = self.context.people()

        unique_values = set()

        for value in (getattr(brain, column.fields[0]) for brain in people):
            if value is None:
                continue

            if isinstance(value, list):
                map(unique_values.add, value)
            else:
                unique_values.add(value)

        return sorted(unique_values, key=tools.unicode_collate_sortkey())

    def selected_column_value(self, column):
        assert column.selectable

        filter = self.filter

        if isinstance(filter, ListFilter) and filter.key == column.fields[0]:
            return filter.value.encode('utf-8')
        else:
            return '__all__'



import codecs

from copy import copy
from five import grok
from importlib import import_module
from plone import api
from plone.directives import form
from z3c.form import field
from z3c.form.interfaces import HIDDEN_MODE
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from z3c.form.browser.radio import RadioFieldWidget
from zope import schema
from zope.component import getAdapters
from zope.i18nmessageid import MessageFactory
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from seantis.people import _
from seantis.people import catalog_id
from seantis.people.interfaces import IList, IExportVariant
from seantis.people.browser import BaseForm
from seantis.people.errors import ContentExportError
from seantis.people.content.export_content import (
    supported_formats, export_people
)

uppercase_formats = [f.upper() for f in supported_formats]


class IExportFormSchema(form.Schema):

    export_fields = schema.List(
        title=_(u"Fields"),
        description=_(u"Fields to include in the export"),
        required=True,
        value_type=schema.Choice(values=[]),
    )

    export_format = schema.Choice(
        title=_(u"Format"),
        values=uppercase_formats,
        default='CSV'
    )

    variant = schema.Choice(
        title=_(u"Variant"),
        values=[]
    )

    review_state = schema.Choice(
        title=_(u"State"),
        values=[]
    )

    include_inactive = schema.Bool(
        title=_(u"Include Inactive"),
        default=True
    )


class ExportForm(BaseForm):

    grok.require('cmf.ModifyPortalContent')
    grok.context(IList)
    grok.name('export')

    ignoreContext = True
    output = None
    enable_unload_protection = False

    def render(self):
        if self.output:
            return self.output
        else:
            return super(ExportForm, self).render()

    @property
    def label(self):
        return _(u"Export ${name}", mapping={
            'name': self.context.title
        })

    @property
    def portal_type(self):
        return self.context.used_type()

    @property
    def portal_type_fields(self):
        result = []

        if self.portal_type:
            schema = self.portal_type.lookupSchema()

            for id, f in field.Fields(schema).items():
                result.append((id, f.field.title))

            module = '.'.join(self.portal_type.klass.split('.')[:-1])
            klass = self.portal_type.klass.split('.')[-1]
            klass = getattr(import_module(module), klass)

            membership_fields = klass().membership_fields
            for id in sorted(membership_fields):
                result.append((id, membership_fields[id]))

        return result

    @property
    def fields(self):
        fields = field.Fields(IExportFormSchema)
        self.prepare_export_fields(fields)
        self.prepare_format_field(fields)
        self.prepare_review_states_field(fields)
        self.prepare_variants(fields)
        return fields

    @property
    def hide_include_inactive(self):
        """ The include inactive flag is only shown if any people have it
        set. If it wouldn't make a difference, don't include it.

        """
        path = {
            'query': '/'.join(self.context.getPhysicalPath()), 'depth': 1
        }
        catalog = api.portal.get_tool(catalog_id)
        return len(catalog(path=path, is_active_person=False)) == 0

    def updateWidgets(self):
        super(ExportForm, self).updateWidgets()

        if self.hide_include_inactive:
            self.widgets['include_inactive'].mode = HIDDEN_MODE

        if len(self.available_variants) == 1:
            self.widgets['variant'].mode = HIDDEN_MODE

    def prepare_export_fields(self, fields):

        vocabulary = SimpleVocabulary(terms=[
            SimpleTerm(id, title=title)
            for id, title in self.portal_type_fields
        ])

        default_values = [i[0] for i in self.portal_type_fields]

        f = fields['export_fields']
        f.field = copy(f.field)

        f.field.value_type.vocabulary = vocabulary
        f.field.default = default_values

        f.widgetFactory = CheckBoxFieldWidget

    def prepare_format_field(self, fields):
        f = fields['export_format']
        f.field = copy(f.field)
        f.widgetFactory = RadioFieldWidget

    def prepare_review_states_field(self, fields):

        vocabulary = SimpleVocabulary(terms=[
            SimpleTerm(id, title=title)
            for id, title in self.available_review_states
        ])

        f = fields['review_state']
        f.field = copy(f.field)

        f.field.vocabulary = vocabulary
        f.field.default = '__all__'

        f.widgetFactory = RadioFieldWidget

    def prepare_variants(self, fields):

        vocabulary = SimpleVocabulary(terms=[
            SimpleTerm(name, title=self.translate(title))
            for name, title in self.available_variants
        ])

        f = fields['variant']
        f.field = copy(f.field)

        f.field.vocabulary = vocabulary
        f.field.default = '__none__'

        f.widgetFactory = RadioFieldWidget

    @property
    def available_review_states(self):
        all_states = [
            ('__all__', _(u'All'))
        ]

        used_type = self.context.used_type()
        if not used_type:
            return all_states

        actual_states = set(b.review_state for b in self.context.people())
        wftool = api.portal.get_tool('portal_workflow')

        _plone = MessageFactory('plone')

        for state in actual_states:
            all_states.append((
                state, _plone(
                    wftool.getTitleForStateOnType(state, used_type.id)
                )
            ))

        return all_states

    @property
    def available_variants(self):
        exports = [
            ('__none__', _(u'Standard'))
        ]

        for name, export in getAdapters((self.context, ), IExportVariant):
            if export.can_handle_type(self.portal_type):
                exports.append((name, export.title))

        return exports

    def get_variant_by_name(self, name):
        if name == '__none__':
            return None

        for variant, export in getAdapters((self.context, ), IExportVariant):
            if variant == name and export.can_handle_type(self.portal_type):
                return export

        return None

    @property
    def available_actions(self):
        yield dict(name='export', title=_(u"Export"),
                   css_class='context allowMultiSubmit')
        yield dict(name='cancel', title=_(u"Cancel"))

    def handle_export(self):
        try:
            export_fields = self.parameters.get('export_fields')
            export_fields = [
                (id, title) for id, title
                in self.portal_type_fields
                if id in export_fields
            ]

            review_state = self.parameters.get('review_state')
            review_state = None if review_state == '__all__' else review_state

            include_inactive = self.parameters.get('include_inactive')

            dataset = export_people(
                self.request,
                self.context,
                self.portal_type.id,
                export_fields,
                review_state,
                include_inactive
            )

            variant = self.get_variant_by_name(self.parameters.get('variant'))
            if variant:
                dataset = variant.adjust_dataset(dataset, self.parameters)

                if dataset is None:
                    raise ContentExportError(
                        _(u"The selected variant could not be applied")
                    )

            format = self.parameters.get('export_format').lower()
            filename = '%s.%s' % (self.context.title, format)
            filename = codecs.utf_8_encode('filename="%s"' % filename)[0]

            output = getattr(dataset, format)

            RESPONSE = self.request.RESPONSE
            RESPONSE.setHeader("Content-disposition", filename)
            RESPONSE.setHeader(
                "Content-Type", "application/%s;charset=utf-8" % format
            )
            RESPONSE.setHeader("Content-Length", len(output))

            self.output = output

        except ContentExportError as e:
            self.raise_action_error(e.translate(self.request))



from zope.security import checkPermission as has_permission
from plone.autoform.interfaces import READ_PERMISSIONS_KEY


def get_read_permissions(schema):
    return schema.queryTaggedValue(READ_PERMISSIONS_KEY, {})


def set_read_permissions(schema, permissions):
    return schema.setTaggedValue(READ_PERMISSIONS_KEY, permissions)


def has_read_access(schema, field, context):

    required_permission = get_read_permissions(schema).get(field)

    if required_permission:
        return has_permission(required_permission, context)
    else:
        return True



from plone.supermodel import loadString

from seantis.people.supermodel import security
from seantis.people import tests


class TestSecurity(tests.IntegrationTestCase):

    def test_read_permissions(self):
        model = loadString("""
            <model xmlns="http://namespaces.plone.org/supermodel/schema">
                <schema />
            </model>
        """)

        security.set_read_permissions(
            model.schema, {'fieldname': 'zope2.View'}
        )
        self.assertEqual(security.get_read_permissions(model.schema), {
            'fieldname': 'zope2.View'
        })

    def test_had_read_access(self):
        model = loadString("""
            <model xmlns="http://namespaces.plone.org/supermodel/schema">
                <schema />
            </model>
        """)

        context = self.new_temporary_folder()

        # no permission -> has access
        self.assertTrue(
            security.has_read_access(model.schema, 'fieldname', context)
        )

        security.set_read_permissions(
            model.schema, {'fieldname': 'cmf.ManagePortal'}
        )

        # anonymous has no access now
        self.assertFalse(
            security.has_read_access(model.schema, 'fieldname', context)
        )

        # but the admin does
        with self.user('admin'):
            self.assertTrue(
                security.has_read_access(model.schema, 'fieldname', context)
            )



from mock import patch, Mock

from z3c.form.interfaces import ActionExecutionError

from seantis.people import tests
from seantis.people.interfaces import IPerson


class MockFile(object):
    def __init__(self, filename, data):
        self.filename = filename
        self.data = data


class TestImportForm(tests.IntegrationTestCase):

    def test_import_form_fields(self):
        self.login('admin')

        new_type = self.new_temporary_type(
            title='employee',
            behaviors=[IPerson.__identifier__]
        )

        form = self.get_import_form()
        types = form.fields['portal_type'].field.vocabulary.by_token

        self.assertEqual(types.values()[0].title, u'employee')
        self.assertEqual(types.values()[0].value, new_type.id)
        self.assertEqual(types.values()[0].token, new_type.id)

    def test_import_format(self):
        form = self.get_import_form()

        self.assertEqual(form.get_format_from_filename('asdf'), '')
        self.assertEqual(form.get_format_from_filename('file.txt'), 'txt')
        self.assertEqual(form.get_format_from_filename('file.xls.bak'), 'bak')

    @patch('transaction.abort')
    def test_unsupported_format(self, abort):
        form = self.get_import_form()

        form.extractData = Mock(return_value=({
            'portal_type': '',
            'import_file': MockFile('test.jpg', None)
        }, None))

        self.assertRaises(ActionExecutionError, form.handle_import)
        self.assertTrue(abort.called)

    @patch('seantis.people.browser.import_form.import_people')
    def test_run_import(self, import_people):
        form = self.get_import_form()

        form.extractData = Mock(return_value=({
            'portal_type': '',
            'import_file': MockFile('test.csv', None)
        }, None))

        import_people.return_value = 3

        form.handle_import()

        self.assertEqual(form.status.mapping['count'], 3)



import json

from lxml import etree

from plone.autoform.widgets import WidgetExportImportHandler
from plone.supermodel.utils import noNS


class JsonWidgetExportImportHandler(WidgetExportImportHandler):
    """ plone.autoform widget attributes handler. May be given an arbitrary
    amount of attributes used on a widget, together with the expected type.

    The values are then parsed and stored as JSON.

    For example, the following xml...

    <field name="birthday" type="zope.schema.Date">
      <form:widget type="plone.formwidget.datetime.z3cform.widget.DateWidget">
        <years_range>[-100, 10]</years_range>
      </form:widget>
    </field>

    ...together with the following setup...

    Handler = JsonWidgetExportImportHandler({
        'years_range': tuple
    })

    ...will ensure that the years_range in json is read from the xml and used
    as widget parameter in json.

    """

    def __init__(self, attributes):
        self.attributes = attributes

    def read(self, widget_node, params):

        # get all child node attributes in one swoop
        child_nodes = dict(
            (noNS(c.tag), c.text.strip()) for c in widget_node.iterchildren()
            if noNS(c.tag) in self.attributes
        )

        for attr, value in child_nodes.items():
            params[attr] = self.attributes[attr](json.loads(value))

    def write(self, widget_node, params):
        for attribute in self.attributes:
            if params.get(attribute) is not None:
                child = etree.Element(attribute)
                child.text = json.dumps(params.get(attribute))
                widget_node.append(child)


DateWidgetExportImportHandler = JsonWidgetExportImportHandler({
    'years_range': tuple
})



import os, os.path

from five import grok
from plone import api
from Products.CMFPlone.interfaces import IPloneSiteRoot

from seantis.people.browser.base import BaseView
from seantis.people.interfaces import IPerson


class ControlPanel(BaseView):

    permission = 'cmf.ManagePortal'
    grok.require(permission)
    grok.context(IPloneSiteRoot)
    grok.name('seantis-people-controlpanel')

    template = grok.PageTemplateFile('templates/controlpanel.pt')

    @property
    def setup(self):
        return api.portal.get_tool('portal_setup')

    @property
    def installer(self):
        return api.portal.get_tool('portal_quickinstaller')

    def update(self):
        if 'upgrade-profile' in self.request:
            profile = self.get_profile_by_id(self.request['upgrade-profile'])

            if profile:
                self.upgrade_profile(profile)

        if 'install-profile' in self.request:
            profile = self.get_profile_by_id(self.request['install-profile'])

            if profile:
                self.install_profile(profile)

    def upgrade_profile(self, profile):
        # copied from Products.CMFCore.QuickInstallerTool
        for upgrade in self.setup.listUpgrades(profile['id']):
            if not type(upgrade) is list:
                upgrade = [upgrade]
            for upgradestep in upgrade:
                step = upgradestep['step']
                step.doStep(self.setup)
        version = str(profile['version'])
        if version == 'latest':
            version = self.installer.getLatestUpgradeStep(profile['id'])
        self.set_last_version(profile, version)

    def install_profile(self, profile):
        pid = 'profile-{}'.format(profile['id'])
        self.setup.runAllImportStepsFromProfile(pid)
        self.set_last_version(profile)

    def get_profile_by_id(self, profile_id):
        profiles = self.relevant_profiles()

        for profile in profiles:
            if profile['id'] == profile_id:
                return profile

        return None

    def get_last_version(self, profile):
        version = self.setup.getLastVersionForProfile(profile['id'])

        if version == 'unknown':
            return None
        else:
            return version[0]

    def set_last_version(self, profile, version=None):
        version = version or profile['version']

        self.setup.setLastVersionForProfile(profile['id'], version)

    def relevant_profiles(self):
        return [
            p for p in self.setup.listProfileInfo() if self.has_people_type(p)
        ]

    def has_people_type(self, profile):
        try:
            types = os.path.join(profile['path'], 'types')
            files = os.listdir(types)
        except OSError:
            return False

        for f in files:
            try:
                with open(os.path.join(types, f), 'r') as typeinfo:
                    for line in typeinfo:
                        if IPerson.__identifier__ in line:
                            return True
            except IOError:
                continue

        return False

    def is_installed_profile(self, profile):
        return self.get_last_version(profile) is not None

    def is_latest_profile(self, profile):
        return self.get_last_version(profile) == profile['version']

    def available_profiles(self):
        available = lambda p: not self.is_installed_profile(p)
        return [p for p in self.relevant_profiles() if available(p)]

    def installed_profiles(self):
        installed = lambda p: self.is_installed_profile(p)
        return [p for p in self.relevant_profiles() if installed(p)]

    def upgradeable_profiles(self):
        upgradeable = lambda p: not self.is_latest_profile(p)
        return [p for p in self.installed_profiles() if upgradeable(p)]



from plone.app.layout.globals.layout import LayoutPolicy

from seantis.people.interfaces import IList


class ListLayoutPolicy(LayoutPolicy):
    """ Adds the currently used type to the list view as body class."""

    def bodyClass(self, template, view):
        """Returns the CSS class to be used on the body tag.
        """
        if IList.providedBy(self.context):
            used_type = self.context.used_type()
        else:
            used_type = None

        body_class = LayoutPolicy.bodyClass(self, template, view)

        if used_type is not None:
            # e.g. seantis.people.standard becomes seantis-people-standard-list
            type_class = '{}-list'.format(used_type.id.replace('.', '-'))
            return ' '.join((body_class, type_class))
        else:
            return body_class



from seantis.people import tests


class TestBrowser(tests.FunctionalTestCase):

    def setUp(self):
        super(TestBrowser, self).setUp()
        self.baseurl = self.portal.absolute_url()
        self.admin_browser = browser = self.new_admin_browser()

        browser.open(self.baseurl + '/createObject?type_name=Folder')
        browser.getControl('Title').value = 'testfolder'
        browser.getControl('Save').click()

        self.folder_url = self.baseurl + '/testfolder'
        self.infolder = lambda url: self.folder_url + url

    def tearDown(self):
        self.admin_browser.open(self.infolder('/delete_confirmation'))
        self.admin_browser.getControl('Delete').click()

        super(TestBrowser, self).tearDown()

    def new_admin_browser(self):
        browser = self.new_browser()
        browser.login_admin()

        return browser

    def new_people_list(self, title='testlist'):
        browser = self.new_admin_browser()

        browser.open(self.infolder('/++add++seantis.people.list'))
        browser.getControl(name='form.widgets.title').value = title
        browser.getControl('Save').click()

        return self.infolder('/{}'.format(title.lower()))

    def new_person(self, first_name, last_name, list_id='testlist'):
        browser = self.new_admin_browser()

        browser.open(
            self.infolder('/{}/++add++seantis.people.standard'.format(list_id))
        )

        browser.getControl('First Name').value = first_name
        browser.getControl('Last Name').value = last_name

        browser.getControl('Save').click()

    def test_empty_list(self):
        list_url = self.new_people_list()

        browser = self.new_admin_browser()
        browser.open(list_url)
        self.assertIn('No people in the list.', browser.contents)

        # layout policy
        self.assertNotIn('seantis-people-standard-list', browser.contents)

    def test_simple_list(self):
        list_url = self.new_people_list()
        self.new_person('Phil', 'Dunphy')

        browser = self.new_admin_browser()
        browser.open(list_url)
        self.assertIn('Phil', browser.contents)
        self.assertIn('Dunphy', browser.contents)

        # layout policy
        self.assertIn('seantis-people-standard-list', browser.contents)

    def test_remove_from_list(self):
        list_url = self.new_people_list()
        self.new_person('Claire', 'Dunphy')

        browser = self.new_admin_browser()
        browser.open(list_url)
        self.assertIn('Claire', browser.contents)
        self.assertIn('Dunphy', browser.contents)

        browser.open(list_url + '/dunphy-claire/delete_confirmation')
        browser.getControl('Delete').click()

        browser.open(list_url)
        self.assertNotIn('Claire', browser.contents)
        self.assertNotIn('Dunphy', browser.contents)



import logging
log = logging.getLogger('seantis.people')

from five import grok
from plone import api
from ZPublisher.interfaces import IPubBeforeCommit

from seantis.people import catalog_id


@grok.subscribe(IPubBeforeCommit)
def reindex_lists(event):
    """ seantis.people dynamically creates indexes and metadata to keep the
    list of people speedy. Depending on the type this may include an index
    of the organisation-memberships, which are available on the person and
    looked up using the catalog.

    During a clean & rebuild operation these memberships cannot be looked up
    because the index at this point may or may not contain the correct
    memberships (depending on the order of the content tree).

    To work around this issue, the catalog rebuild url is caught here. This
    url will be showing up in this method after the reindex happened.

    We then reindex just the organisations once more, knowing this time
    that the information in the catalog is up to date.

    """

    # never fail, this code is run after ever request!
    try:
        if event.request.URL.endswith('manage_catalogRebuild'):
            catalog = api.portal.get_tool(catalog_id)

            lists = catalog.unrestrictedSearchResults(
                portal_type='seantis.people.list'
            )

            for brain in lists:
                brain.getObject().reindex_organisations()

            log.info("organisation reindex successful")
    except Exception:
        log.exception("exception thrown during organisation reindex")



import logging
log = logging.getLogger('seantis.people')

from urllib import urlopen
from collections import namedtuple

import tablib

from plone import api
from plone import namedfile

from zope.schema import getFields, getValidationErrors, Text
from zope.schema._bootstrapinterfaces import RequiredMissing
from zope.schema.interfaces import (
    IFromUnicode, ValidationError, SchemaNotFullyImplemented,
    IDate, IDatetime, IURI, ITextLine, IList, ISet, IChoice
)

from isodate import parse_date, parse_datetime

from seantis.people import _
from seantis.people.errors import ContentImportError
from seantis.plonetools import tools

supported_formats = ('csv', )


def import_people(request, container, portal_type, format, data):
    dataset = get_dataset(format, data)
    attribute_map = get_attribute_map(request, dataset.headers, portal_type)

    schema = tools.get_schema_from_portal_type(portal_type)

    for ix, record in enumerate(dataset.dict):
        log.info('processing row number {}'.format(ix+1))
        try:
            values = get_attribute_values(record, attribute_map)

            # add None for missing values
            for field in getFields(schema).keys():
                if field not in values:
                    values[field] = None

            validate_attribute_values(schema, values)

            api.content.create(
                container=container,
                type=portal_type,
                id='',
                **values
            )
        except ContentImportError, e:
            raise ContentImportError(
                e.message, rownumber=ix+1, colname=e.colname
            )
        except Exception, e:
            log.exception('error importing people')
            raise ContentImportError(e.message, rownumber=ix+1)

    return ix + 1


def get_dataset(format, data):
    assert format in supported_formats

    def get_dataset_from_csv():
        ds = tablib.Dataset()
        ds.csv = data

        return ds

    return locals()['get_dataset_from_{}'.format(format)]()


def get_attribute_map(request, headers, portal_type):

    if not headers:
        raise ContentImportError(_(u'No column headers were found'))

    schema = tools.get_schema_from_portal_type(portal_type)
    attribute_map = {}

    fields = getFields(schema)

    known_titles = dict((field.title, field) for field in fields.values())
    known_fields = dict((key, field) for key, field in fields.items())

    translate = tools.translator(request)
    known_translated_titles = dict(
        (translate(field.title), field) for field in fields.values()
    )

    for header in headers:

        if header in known_translated_titles:
            field = known_translated_titles[header]
        elif header in known_fields:
            field = known_fields[header]
        elif header in known_titles:
            field = known_titles[header]
        else:
            log.warn(u"The header '{}' was ignored.".format(header))
            continue

        if field in attribute_map.values():
            raise ContentImportError(
                _(
                    u'The ${name} column is specified more than once',
                    mapping=dict(name=header)
                )
            )

        if field in ('title', 'id'):
            raise ContentImportError(
                _(
                    u'The ${name} column is invalid. "title" and "id" are '
                    u'reserved fieldnames which may not be used.',
                    mapping=dict(name=header)
                )
            )

        attribute_map[header] = field

    return attribute_map


def download_field_from_url(field, url):
    """ Download the field from the given url if supported by the field.
    Returns False if the download does not apply to the given field. If None
    is returned, the value of the field actually is None.

    """

    downloadables = {
        namedfile.field.NamedImage: field._type,
        namedfile.field.NamedBlobImage: field._type
    }

    url = url.strip()

    # images are fetched through their url if possible
    for fieldtype, klass in downloadables.items():
        if isinstance(field, fieldtype):
            try:
                image = klass(urlopen(url).read())
                if -1 in image.getImageSize():
                    return None
                else:
                    return image
            except IOError:
                return None

    return False


def convert_to_list(value):
    if '\n' in value:
        separator = '\n'
    else:
        separator = ','

    return [s.strip() for s in value.split(separator)]


def get_attribute_values(record, attribute_map):
    values = {}

    for header, field in attribute_map.items():

        downloaded = download_field_from_url(field, record[header])
        if downloaded is not False:
            values[field.__name__] = downloaded
            continue

        if IDate.providedBy(field):
            if not record[header]:
                values[field.__name__] = None
            else:
                values[field.__name__] = parse_date(record[header])
            continue

        if IDatetime.providedBy(field):
            if not record[header]:
                values[field.__name__] = None
            else:
                values[field.__name__] = parse_datetime(record[header])
            continue

        if IURI.providedBy(field):
            if not record[header].strip():
                values[field.__name__] = None
                continue

        if IList.providedBy(field):
            if ITextLine.providedBy(field.value_type):
                values[field.__name__] = convert_to_list(record[header])
                continue

        if ISet.providedBy(field):
            if IChoice.providedBy(field.value_type):
                values[field.__name__] = set(convert_to_list(record[header]))
                continue

        assert IFromUnicode.providedBy(field), """
            {} does not support fromUnicode
        """.format(field)

        try:
            values[field.__name__] = field.fromUnicode(record[header])

            if isinstance(values[field.__name__], basestring):
                values[field.__name__] = values[field.__name__].strip()
            if isinstance(field, Text):
                values[field.__name__] = values[field.__name__].replace(
                    '<br />', '\n'
                )

        except ValidationError, e:
            raise ContentImportError(e.doc(), colname=header)
        except ValueError, e:
            raise ContentImportError(e.message, colname=header)

    return values


def validate_attribute_values(schema, values):
    obj = namedtuple('ImportObject', values.keys())(**values)

    for error in getValidationErrors(schema, obj):
        column = schema[error[0]].title

        if isinstance(error[1], (SchemaNotFullyImplemented, RequiredMissing)):
            raise ContentImportError(
                _(u'Required column is missing'), colname=column
            )
        else:
            raise ContentImportError(error, colname=column)



import json
from collections import namedtuple

from five import grok

from seantis.people.interfaces import IList
from seantis.people.browser import BaseView


JsonField = namedtuple('JsonField', ['scope', 'getter'])


class JsonListView(BaseView):

    permission = 'zope2.View'
    grok.require(permission)
    grok.context(IList)
    grok.name('json')

    available_fields = {
        'uuid': lambda b: b.UID,
        'title': lambda b: b.Title,
        'url': lambda b: b.getURL()
    }

    def render(self):
        fields = self.available_fields.items()

        records = []

        for brain in self.context.people():
            records.append(dict((field, fn(brain)) for field, fn in fields))

        self.request.response.setHeader('Content-type', 'application/json')
        return json.dumps(records)



from plone import api
from plone.dexterity.content import Container
from seantis.people.interfaces import IPerson
from seantis.people import _
from seantis.people import catalog_id


class PersonBase(Container):

    def _getCatalogTool(self):
        try:
            return api.portal.get_tool(catalog_id)
        except (api.exc.CannotGetPortalError, api.exc.InvalidParameterError):
            # it's possible that this is called during an update step, at
            # which point the seantis people catalog is not yet available
            return None

    @property
    def memberships(self):
        return IPerson(self).memberships()

    @property
    def organizations(self):
        if hasattr(self, 'test_organizations'):
            return self.test_organizations

        return IPerson(self).organizations()

    @property
    def organization_uuids(self):
        if hasattr(self, 'test_organization_uuids'):
            return self.test_organization_uuids

        return IPerson(self).organization_uuids()

    @property
    def membership_fields(self):
        return {
            'memberships': _(u'Memberships')
        }



from seantis.people import catalog_id
from seantis.plonetools import setuphandlers

indexes = [
    ('first_letter', 'FieldIndex'),
    ('membership_person', 'FieldIndex'),
    ('is_active_person', 'FieldIndex')
]


def add_catalog_indexes(context, logger=None):
    # these indexes are not dynamic and need to be available in both catalogs
    # -> only the dynamic indexes created by the supermodel are only bound
    # to the people's catalog
    for catalog in ('portal_catalog', catalog_id):
        setuphandlers.add_catalog_indexes(
            'seantis.people', indexes, context, logger, catalog
        )


def import_indexes(context):
    # see above..
    for catalog in ('portal_catalog', catalog_id):
        setuphandlers.import_indexes(
            'seantis.people', indexes, context, catalog
        )



from zope.component import getAdapters
from plone import api

from seantis.plonetools import tools
from seantis.people.interfaces import ICompoundColumns
from seantis.people.supermodel.security import has_read_access
from seantis.people.supermodel import (
    get_columns,
    get_custom_column_titles,
    get_selectable_fields,
    get_title_fields,
)


# Compound columns are columns where the key is used for filtering and the
# value is used for display. For example, the organizations column contains
# the name of the organizations a person belongs to, but to render these
# in the list, the organization_uuids attribute is used which being rendered
# with links to the actual organizations.
def get_compound_columns():
    columns = {
        'organizations': 'organization_uuids'
    }
    for name, adapter in getAdapters((api.portal.get(), ), ICompoundColumns):
        columns.update(adapter.get_compound_columns())

    return columns


def unrestricted_get_schema_columns(schema):
    return _get_schema_columns(schema, None, restricted=False)


def get_schema_columns(schema, context):
    return _get_schema_columns(schema, context, restricted=True)


def _get_schema_columns(schema, context, restricted):
    columns = []
    titles = get_custom_column_titles(schema)

    for ix, fields in enumerate(get_columns(schema)):
        if restricted:
            fields = [
                f for f in fields if has_read_access(schema, f, context)
            ]

        if fields:
            columns.append(SchemaColumn(schema, fields, titles[ix]))

    return columns


class SchemaColumn(object):

    def __init__(self, schema, fields, custom_title=None):
        self.schema = schema
        self.fields = tools.order_fields_by_schema(fields, schema)

        if custom_title is not None:
            self.titles = [custom_title]
        else:
            self.titles = []
            for field in self.fields:
                if field in self.schema:
                    self.titles.append(self.schema[field].title)
                else:
                    self.titles.append(field)

        self.contains_title_field = bool(
            set(self.fields) & set(get_title_fields(self.schema))
        )
        self.selectable = (
            len(self.fields) == 1 and self.fields[0] in get_selectable_fields(
                self.schema
            )
        )
        self.filter_key = self.selectable and self.fields[0] or None



from plone import api

from seantis.people import tests


class TestControlpanel(tests.IntegrationTestCase):

    @property
    def panel(self):
        with self.user('admin'):
            return api.content.get_view(
                name='seantis-people-controlpanel',
                context=api.portal.get(),
                request=self.request
            )

    def test_install_profiles(self):
        profiles = self.panel.available_profiles()
        map(self.panel.install_profile, profiles)
        
        ids = [p['id'] for p in self.panel.installed_profiles()]
        self.assertEqual(ids, [p['id'] for p in profiles])
        self.assertEqual(self.panel.available_profiles(), [])
        self.assertEqual(self.panel.upgradeable_profiles(), [])

    def test_included_profiles(self):
        # we should see the profiles included with seantis.people
        ids = [p['id'] for p in self.panel.available_profiles()]
        self.assertEqual(sorted(ids), [
            u'seantis.people:phz',
            u'seantis.people:standard'
        ])

        # they are not installed
        ids = [p['id'] for p in self.panel.installed_profiles()]
        self.assertEqual(ids, [])

        # and therefore not upgradable
        ids = [p['id'] for p in self.panel.upgradeable_profiles()]
        self.assertEqual(ids, [])



from mock import Mock

from z3c.form.interfaces import ActionExecutionError

from seantis.people import tests


class TestViewBase(tests.IntegrationTestCase):

    def test_form_parameters(self):
        form = self.get_import_form()

        form.extractData = Mock(return_value=('yo', None))
        self.assertEqual(form.parameters, 'yo')

        form.extractData = Mock(return_value=('yo', 'error'))
        self.assertEqual(form.parameters, None)
        self.assertEqual(form.status, form.formErrorsMessage)

    def test_raise_action_error(self):
        form = self.get_import_form()

        try:
            form.raise_action_error('testmessage')
        except ActionExecutionError, e:
            self.assertEqual(e.error.message, 'testmessage')
        else:
            assert False, "untriggered exception"




from plone.formwidget.contenttree import ObjPathSourceBinder
from plone.directives import form

from z3c.relationfield.schema import RelationChoice

from zope import schema
from zope.interface import Interface, Attribute

from seantis.people import _


class ISeantisPeopleSpecific(Interface):
    pass


class IPersonMarker(Interface):
    pass


class IPerson(Interface):

    def memberships(self):
        """ A dictionary of membership belonging to this person, keyed by
        organization (brain or object with a title and an url)."""


class IPeopleCatalog(Interface):
    """ Seantis People uses a custom ZCatalog implementing this interface. """


class INameFromPerson(Interface):
    pass


class IList(form.Schema):
    """ A list of IPerson objects. """

    title = schema.TextLine(
        title=_(u"Name of the list of people"),
        required=True
    )

    people_per_page = schema.Int(
        title=_(u"Number of people shown on each page"),
        required=True,
        default=50
    )


class IMembershipSource(Interface):

    def memberships(self, person=None):
        """ Return a dictionary of objects providing IMembership. The key
        of the dictionary is the uuid of the organization (object with url and
        title).

        If a person object (not a brain) is given, only the memberships
        of that person are returned.
        """


class IExportVariant(Interface):
    """ An export variant is an adapter that has the ability to intercept the
    export process and change the tablib dataset before it is converted into
    another format (XLS, CSV, JSON...).

    Only one variant is used at a time. The passed dataset may not contain
    all fields that are necesary for the export. In this case 'None' should be
    returned.

    """

    title = Attribute("The title of the export")

    def can_handle_type(self, fti):
        """ Called with the person fti that is about to be exported. Return
        true if that type is handled by this export.

        """
        raise NotImplementedError

    def adjust_dataset(self, dataset, export_options):
        raise NotImplementedError


class IMembershipChangedEvent(Interface):
    """ This event needs to be signaled when the membership in a
    IMembershipSource object changed. That means if a membership was added,
    removed or an attribute it was changed (start, end, role).

    """

    person = Attribute("The person whose membership has changed.")


class IMembership(form.Schema):

    role = schema.TextLine(
        title=_(u"Role inside organization"),
        required=True
    )

    person = RelationChoice(
        title=_(u"Person"),
        source=ObjPathSourceBinder(
            object_provides=IPersonMarker.__identifier__
        ),
        required=True,
    )

    note = schema.TextLine(
        title=_(u"Note"),
        description=_(u"General note about this membership"),
        required=False
    )


class ICompoundColumns(Interface):

    def get_compound_columns(self):
        """ Return a dictionary of compound columns. See
        seantis.people.supermodel.schema_columns for more.

        """



from five import grok

from plone.namedfile.field import NamedBlobImage, NamedImage

from seantis.plonetools import tools
from seantis.people.supermodel import security, get_detail_fields
from seantis.people.interfaces import IPersonMarker
from seantis.people.browser import BaseView, Renderer


class PersonView(BaseView):

    permission = 'zope2.View'
    grok.require(permission)
    grok.context(IPersonMarker)
    grok.name('view')

    template = grok.PageTemplateFile('templates/person.pt')

    types_without_label = (
        NamedBlobImage,
        NamedImage
    )

    def update(self):
        self.renderer = Renderer(self.schema, place='detail')

    @property
    def schema(self):
        return tools.get_schema_from_portal_type(self.context.portal_type)

    def fields(self, position):
        has_access = lambda field: security.has_read_access(
            self.schema, field, self.context
        )
        fields = get_detail_fields(self.schema).get(position)

        if not fields:
            return []

        return [f for f in fields if has_access(f)]

    def get_field_title(self, field):

        # if the object has a custom_titles dictionary it is
        # used to override titles on the detail view
        if hasattr(self.context, 'custom_titles'):
            if field in self.context.custom_titles:
                return self.context.custom_titles[field]

        if field in self.schema:
            if type(self.schema[field]) in self.types_without_label:
                return u''

            return self.schema[field].title
        else:
            return field

    def visible_positions(self):
        positions = ('top', 'left', 'right', 'bottom')
        return [
            p for p in positions if any(
                getattr(self.context, f) for f in self.fields(p)
            )
        ]

    def split_screen(self):
        return set(('left', 'right')).issubset(set(self.visible_positions()))

    def get_parent(self):
        return self.context.aq_inner.aq_parent

    def get_adjacent_urls(self):
        """ Returns the previous and next url of the current item, if found.
        Only works inside people lists.

        """
        parent = self.get_parent()

        if not hasattr(parent, 'people'):
            return None, None

        people = self.get_parent().people()

        if len(people) <= 1:
            return None, None

        people_ids = list(p.id for p in self.get_parent().people())
        context_id = self.context.getId()

        try:
            position = people_ids.index(context_id)
        except ValueError:
            return None, None

        if position == 0:
            return None, people[1].getURL()
        if position == len(people) - 1:
            return people[position - 1].getURL(), None
        else:
            return people[position - 1].getURL(), people[position + 1].getURL()



from five import grok

from seantis.plonetools.browser import BaseView as SharedBaseView
from seantis.plonetools.browser import BaseForm as SharedBaseForm

from seantis.people.interfaces import ISeantisPeopleSpecific


class BaseView(SharedBaseView):

    grok.baseclass()
    grok.layer(ISeantisPeopleSpecific)

    domain = 'seantis.people'


class BaseForm(SharedBaseForm):

    grok.baseclass()
    grok.layer(ISeantisPeopleSpecific)

    domain = 'seantis.people'



# -*- coding: utf-8 -*-

from zope.component import getUtility
from zope.component.hooks import setSite

from plone import api
from plone.dexterity.interfaces import IDexterityFTI

from seantis.people import catalog_id
from seantis.people.supermodel.indexing import update_related_indexes


def import_profile(profile):
    setup = api.portal.get_tool('portal_setup')
    setup.runAllImportStepsFromProfile(profile)


def run_import_step_from_profile(step, module, profile):
    setup = api.portal.get_tool('portal_setup')
    profile = 'profile-{}:{}'.format(module, profile)
    setup.runImportStepFromProfile(profile, step)


def upgrade_portal_type(portal_type, module, profile):
    run_import_step_from_profile('typeinfo', module, profile)
    update_related_indexes(getUtility(IDexterityFTI, portal_type))


def install_custom_controlpanel(context):
    run_import_step_from_profile('controlpanel', 'seantis.people', 'default')


def upgrade_membership_title(context):
    upgrade_portal_type(
        'seantis.people.membership', 'seantis.people', 'default'
    )

    catalog = api.portal.get_tool('portal_catalog')
    memberships = catalog.unrestrictedSearchResults(
        portal_type='seantis.people.membership'
    )

    for membership in [m.getObject() for m in memberships]:
        if not hasattr(membership, 'role') or membership.role is None:
            membership.role = membership.title
            membership.title = membership.get_custom_title()
            membership.reindexObject()

            old_id = membership.id
            safe_title = membership.title.replace(u'ä', u'ae')
            safe_title = safe_title.replace(u'ö', u'oe')
            safe_title = safe_title.replace(u'ü', u'ue')

            new_id = safe_title.encode('ascii', 'ignore')
            new_id = new_id.lower().replace(' ', '-')
            membership.aq_inner.aq_parent.manage_renameObject(old_id, new_id)


def upgrade_phz_type_info(context):
    upgrade_portal_type('seantis.people.phz', 'seantis.people', 'phz')


def upgrade_standard_type_info(context):
    upgrade_portal_type('seantis.people.phz', 'seantis.people', 'standard')


def install_membership_dependencies(context):
    import_profile('profile-plone.formwidget.autocomplete:default')
    import_profile('profile-plone.app.relationfield:default')


def update_people_list(context):
    upgrade_portal_type('seantis.people.list', 'seantis.people', 'default')


def introduce_custom_catalog(context, profiles=None):
    run_import_step_from_profile('toolset', 'seantis.people', 'default')

    import_ctx = context._getImportContext('profile-seantis.people:default')
    old_site = api.portal.get()

    try:
        setSite(import_ctx.getSite())
        catalog = api.portal.get_tool(catalog_id)
        catalog.refreshCatalog(clear=1)
    finally:
        setSite(old_site)



from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from plone import api
from Products.CMFPlone.CatalogTool import CatalogTool
from Products.ZCatalog.ZCatalog import ZCatalog
from Products.ZCTextIndex.ZCTextIndex import PLexicon
from Products.ZCTextIndex.PipelineFactory import element_factory
from seantis.people import catalog_id, log
from seantis.people.interfaces import IPeopleCatalog
from zope.interface import implements


class Extra(object):
    pass


class PeopleCatalog(CatalogTool):
    """ A custom catalog for seantis.people. Required because people may have
    a large number of metadata/indices assigned to them which would slow
    down the rest of the site.

    The catalog is used by all types that inherit from
    seantis.people.types.base.PersonBase. It is supposed to shadow the
    portal_catalog. It does so by applying indexing operations to both itself
    and the portal_catalog.

    This results in a separate catalog that has the same indexes and metadata
    as the portal_catalog plus additional indexes and metadata defined by the
    custom people types.

    This catalog can be used if all that is required is a lookup to people
    indexes/properties - like when displaying the list of people. Anything
    else should still go through portal_catalog.

    Note that the people catalog is automatically update if you work with
    seantis people objects. But you have to use it manually if you need
    to get a custom index/metadata value:

        from plone import api
        from seantis.people import catalog_id

        people_catalog = api.portal.get_tool(catalog_id)

    """

    implements(IPeopleCatalog)

    title = "Seantis People Catalog"
    id = catalog_id
    portal_type = meta_type = 'PeopleCatalog'
    plone_tool = 1

    security = ClassSecurityInfo()
    security.declarePrivate('indexObject')
    security.declarePrivate('reindexObject')
    security.declarePrivate('unindexObject')
    security.declarePrivate('setup_lexicons')
    security.declarePrivate('setup_indexes')
    security.declarePrivate('setup_metadata')

    _properties = (
        {'id': 'title', 'type': 'string', 'mode': 'w'},
    )

    def __init__(self):
        ZCatalog.__init__(self, self.id)

        self.setup_lexicons()
        self.setup_indexes()
        self.setup_metadata()

        # the catalog needs to be cleared after setting up everything
        # or subtle errors will appear (like print failing on brains)
        self._catalog.clear()
        self._catalog.updateBrains()

    @property
    def base_catalog(self):
        return api.portal.get_tool('portal_catalog')

    def indexObject(self, *args, **kwargs):
        self.base_catalog.indexObject(*args, **kwargs)
        super(PeopleCatalog, self).indexObject(*args, **kwargs)

    def reindexObject(self, *args, **kwargs):
        self.base_catalog.reindexObject(*args, **kwargs)
        super(PeopleCatalog, self).reindexObject(*args, **kwargs)

    def unindexObject(self, *args, **kwargs):
        self.base_catalog.unindexObject(*args, **kwargs)
        super(PeopleCatalog, self).unindexObject(*args, **kwargs)

    def setup_lexicons(self):
        # setup the lexicons as some Plone-Catalog internal code depends on it
        # we wouldn't relly have to be so thorough as we don't actually want
        # to support full text searches on the people catalog, but setting
        # up the people catalog as close as possible to the portal catalog
        # ensures that there are no surprises
        lexicons = {
            'plone_lexicon': [
                ('Unicode Whitespace splitter', 'Word Splitter'),
                ('Unicode Ignoring Accents Case Normalizer',
                    'Case Normalizer'),
            ],
            'plaintext_lexicon': [
                ('HTML aware splitter', 'Word Splitter'),
                ('Case Normalizer', 'Case Normalizer'),
                ('Remove listed stop words only', 'Stop Words')
            ],
            'htmltext_lexicon': [
                ('HTML aware splitter', 'Word Splitter'),
                ('Case Normalizer', 'Case Normalizer'),
                ('Remove listed stop words only', 'Stop Words')
            ]
        }

        for lexicon, elements in lexicons.items():
            pipeline = []

            for element in elements:
                element = element_factory.instantiate(element[1], element[0])
                pipeline.append(element)

            plexicon = PLexicon(lexicon)
            plexicon._pipeline = pipeline
            self._setObject(lexicon, plexicon)

    def setup_indexes(self):
        # copy the indexes from the base catalog, so the code using the
        # people catalog doesn't have to know where each index/metadata-column
        # is stored exactly

        # only support known index types
        known_index_types = set((
            'BooleanIndex',
            'DateIndex',
            'DateRangeIndex',
            'ExtendedPathIndex',
            'FieldIndex',
            'GopipIndex',
            'KeywordIndex',
            'UUIDIndex',
            'ZCTextIndex',
        ))

        for index in self.base_catalog.index_objects():
            if index.meta_type not in known_index_types:
                log.warn('Unknown index type: {}'.format(index.meta_type))
                continue

            extra = Extra()

            if hasattr(index, 'lexicon_id'):
                extra.lexicon_id = index.lexicon_id

            if hasattr(index, '_index_type'):
                extra.index_type = index._index_type

            if hasattr(index, 'indexed_attrs'):
                extra.indexed_attrs = index.indexed_attrs

            if hasattr(index, 'since_field'):
                extra.since_field = index.since_field

            if hasattr(index, 'since_field'):
                extra.since_field = index.since_field

            if hasattr(index, 'until_field'):
                extra.until_field = index.until_field

            self.addIndex(index.id, index.meta_type, extra=extra)

            if hasattr(index, 'index_naive_time_as_local'):
                new = self._catalog.getIndex(index.id)
                new.index_naive_time_as_local = index.index_naive_time_as_local

    def setup_metadata(self):
        # copy the metadata from the base catalog
        for name in self.base_catalog._catalog.names:
            self.addColumn(name)


InitializeClass(PeopleCatalog)



from .schemahandler import (
    get_title_fields,
    set_title_fields,
    get_selectable_fields,
    set_selectable_fields,
    get_columns,
    set_columns,
    get_order,
    set_order,
    get_detail_fields,
    set_detail_fields,
    get_custom_column_titles,
    set_custom_column_titles,
    get_list_render_options,
    set_list_render_options,
    get_detail_render_options,
    set_detail_render_options
)

from .schema_columns import (
    get_schema_columns,
    unrestricted_get_schema_columns,
    get_compound_columns
)

from .indexing import (
    get_selectable_field_ix
)


from plone.dexterity.content import Container


class Membership(Container):

    def get_custom_title(self):
        person = self.person and self.person.to_object

        # this happens if the person is deleted along with its memberships
        # https://github.com/seantis/seantis.people/issues/27
        if not person:
            return self.role

        return u' '.join((self.role, person.title))



from textwrap import dedent

from zope import schema
from plone.supermodel import loadString

from seantis.people import tests
from seantis.people import interfaces
from seantis.people.errors import ContentImportError
from seantis.people.content.import_content import (
    get_dataset,
    get_attribute_map,
    get_attribute_values,
    import_people,
    validate_attribute_values
)


class TestImportContent(tests.IntegrationTestCase):

    def test_import_people_csv(self):
        folder = self.new_temporary_folder()
        model = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:title>
                    <people:item>firstname</people:item>
                </people:title>
                <field name="firstname" type="zope.schema.TextLine">
                    <title>First Name</title>
                </field>
                <field name="lastname" type="zope.schema.TextLine">
                    <title>Last Name</title>
                </field>
            </schema>
        </model>"""

        portal_type = self.new_temporary_type(
            model_source=model,
            behaviors=[interfaces.INameFromPerson.__identifier__],
        ).id

        with self.user('admin'):
            import_people(self.request, folder, portal_type, 'csv', dedent("""
                First Name,Last Name,Town
                Peter,Griffin ,Quahog
                Glenn,  Quagmire   ,Quahog
            """.lstrip('\n')))

        self.assertEqual(folder['peter'].firstname, 'Peter')
        self.assertEqual(folder['peter'].lastname, 'Griffin')
        self.assertEqual(folder['glenn'].firstname, 'Glenn')
        self.assertEqual(folder['glenn'].lastname, 'Quagmire')

        # no town in the schema
        self.assertFalse(hasattr(folder['peter'], 'town'))
        self.assertFalse(hasattr(folder['glenn'], 'town'))

    def test_import_people_error_in_row(self):
        folder = self.new_temporary_folder()
        model = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:title>
                    <people:item>name</people:item>
                </people:title>
                <field name="name" type="zope.schema.TextLine">
                    <title>Name</title>
                </field>
                <field name="age" type="zope.schema.Int">
                    <title>Age</title>
                </field>
            </schema>
        </model>"""

        portal_type = self.new_temporary_type(
            model_source=model,
            behaviors=[interfaces.INameFromPerson.__identifier__],
        ).id

        self.login('admin')

        # error with a column available
        try:
            import_people(self.request, folder, portal_type, 'csv', dedent("""
                Name,Age
                Macallan,15
                Glenmorangie,Sixteen
            """.lstrip('\n')))
        except ContentImportError, e:
            self.assertIn('invalid literal for int', e.message)
            self.assertEqual(e.rownumber, 2)
            self.assertEqual(e.colname, 'Age')
        else:
            assert False, "The exception should have occurred."

        # error without column available
        try:
            # no folder is given
            import_people(self.request, None, portal_type, 'csv', dedent("""
                Name,Age
                Macallan,15
            """.lstrip('\n')))
        except ContentImportError, e:
            self.assertIn('Missing required parameter', e.message)
            self.assertEqual(e.rownumber, 1)
            self.assertEqual(e.colname, None)
        else:
            assert False, "The exception should have occurred."

    def test_import_people_schema_validation(self):
        model = loadString("""<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:title>
                    <people:item>name</people:item>
                </people:title>
                <field name="name" type="zope.schema.TextLine">
                    <title>Name</title>
                </field>
                <field name="age" type="zope.schema.Int" required="false">
                    <title>Age</title>
                    <required>False</required>
                </field>
            </schema>
        </model>""")

        try:
            validate_attribute_values(model.schema, {'age': 1})
        except ContentImportError, e:
            self.assertIs(type(e), ContentImportError)
            self.assertEqual(e.colname, u'Name')
            self.assertEqual('Required column is missing', e.message)
        else:
            assert False, "The exception should have occurred."

        try:
            validate_attribute_values(model.schema, {'age': None})
        except ContentImportError, e:
            self.assertIs(type(e), ContentImportError)
            self.assertEqual(e.colname, u'Name')
            self.assertEqual('Required column is missing', e.message)
        else:
            assert False, "The exception should have occurred."

        validate_attribute_values(model.schema, {'name': u'one', 'age': 1})

    def test_get_attribute_values(self):
        casts = [
            (schema.Text, u'Test', u'Test'),
            (schema.Int, u'1', 1),
            (schema.Bool, u'True', True),
            (schema.Bool, u'False', False),
            (schema.Bool, u'True', True),
            (schema.Bool, u'', False)
        ]
        for ix, (fieldtype, value, expected) in enumerate(casts):
            key = str(ix)

            record = {key: value}
            attrmap = {key: fieldtype(__name__=key)}

            result = get_attribute_values(record, attrmap)
            self.assertEqual(result[key], expected)

    def test_get_attribute_values_validation_error(self):
        record = {'test': u'www.example.com'}
        attrmap = {'test': schema.URI(__name__='test')}

        try:
            get_attribute_values(record, attrmap)
        except ContentImportError, e:
            self.assertEqual(e.colname, 'test')
            self.assertEqual(e.message, u'The specified URI is not valid.')
        else:
            assert False, "The exception should have occured."

    def test_get_csv_dataset(self):
        ds = get_dataset('csv', dedent("""
            First Name,Last Name,Town
            Peter,Griffin,Quahog
            Glenn,Quagmire,Quahog
        """.lstrip('\n')))

        self.assertEqual(ds.headers, ['First Name', 'Last Name', 'Town'])
        self.assertEqual(ds.dict[0], {
            'First Name': 'Peter',
            'Last Name': 'Griffin',
            'Town': 'Quahog'
        })
        self.assertEqual(ds.dict[1], {
            'First Name': 'Glenn',
            'Last Name': 'Quagmire',
            'Town': 'Quahog'
        })

    def test_get_attribute_map(self):
        model = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema">
            <schema>
                <field name="firstname" type="zope.schema.TextLine">
                    <title>First Name</title>
                </field>
                <field name="lastname" type="zope.schema.TextLine">
                    <title>Last Name</title>
                </field>
            </schema>
        </model>"""

        portal_type = self.new_temporary_type(model_source=model).id

        self.assertRaises(
            ContentImportError,
            get_attribute_map, self.request, [], portal_type
        )

        try:
            get_attribute_map(
                self.request, ['firstname', 'First Name'], portal_type
            )
        except ContentImportError, e:
            self.assertIn('column is specified more than once', e.message)
        else:
            assert False, "The exception should have occured."

        attrmap = get_attribute_map(
            self.request, ['firstname', 'lastname'], portal_type
        )

        self.assertEqual(len(attrmap), 2)
        self.assertEqual(attrmap['firstname'].__name__, 'firstname')
        self.assertEqual(attrmap['lastname'].__name__, 'lastname')

        attrmap = get_attribute_map(
            self.request, ['First Name', 'Last Name'], portal_type
        )

        self.assertEqual(len(attrmap), 2)
        self.assertEqual(attrmap['First Name'].__name__, 'firstname')
        self.assertEqual(attrmap['Last Name'].__name__, 'lastname')

        attrmap = get_attribute_map(self.request, ['foo', 'bar'], portal_type)
        self.assertEqual(len(attrmap), 0)



from plone.supermodel import loadString, serializeSchema

from seantis.people import tests


class TestSchemaColumns(tests.IntegrationTestCase):

    test_model = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:form="http://namespaces.plone.org/supermodel/form">
          <schema>
            <field name="birthday" type="zope.schema.Date">
              <title>Go Shorty</title>
              <form:widget
                type="plone.formwidget.datetime.z3cform.widget.DateWidget">
                <years_range>[-100, 10]</years_range>
              </form:widget>
            </field>
          </schema>
        </model>"""

    def test_read_widget(self):
        model = loadString(self.test_model)
        widgets = model.schema.getTaggedValue('plone.autoform.widgets')

        self.assertEqual(
            widgets['birthday'].params, {'years_range': (-100, 10)}
        )

    def test_write_widget(self):
        model = loadString(self.test_model)
        xml = serializeSchema(model.schema)

        self.assertIn('<years_range>[-100, 10]</years_range>', xml)



from zope.interface import implements
from seantis.people.interfaces import IMembershipChangedEvent, IPersonMarker


class MembershipChangedEvent(object):
    implements(IMembershipChangedEvent)

    def __init__(self, person):
        assert IPersonMarker.providedBy(person)
        self.person = person



from plone import api
from plone.uuid.interfaces import IUUID

from seantis.people.interfaces import IPerson
from seantis.people.content import Membership
from seantis.people import tests


class TestMembership(tests.IntegrationTestCase):

    def test_membership_creation(self):
        with self.user('admin'):
            ms = api.content.create(
                id='test',
                type='seantis.people.membership',
                container=self.new_temporary_folder()
            )

        self.assertIs(type(ms.aq_base), Membership)

    def test_organizations_macro(self):
        person_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )

        with self.user('admin'):
            organization = self.new_temporary_folder('S.H.I.E.L.D')

            nick_fury = api.content.create(
                title='Nick Fury',
                type=person_type.id,
                container=self.new_temporary_folder()
            )

            api.content.create(
                title='Director',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury,
                role='Director'
            )

            macros = nick_fury.unrestrictedTraverse('@@seantis-people-macros')
            organizations = macros.organizations(nick_fury, 'memberships')

        self.assertEqual(len(organizations), 1)
        self.assertEqual(organizations[0].title, 'S.H.I.E.L.D')
        self.assertEqual(organizations[0].role, 'Director')

    def test_membership_person_relation(self):
        person_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )

        with self.user('admin'):
            organization = self.new_temporary_folder('S.H.I.E.L.D')

            nick_fury = api.content.create(
                title='Nick Fury',
                type=person_type.id,
                container=self.new_temporary_folder()
            )

            tony_stark = api.content.create(
                title='Tony Stark',
                type=person_type.id,
                container=self.new_temporary_folder()
            )

            api.content.create(
                title='Director',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury
            )

            api.content.create(
                title='Head',
                type='seantis.people.membership',
                container=organization,
                person=tony_stark
            )

        memberships = IPerson(nick_fury).memberships()

        org = IUUID(organization)

        self.assertEqual(len(memberships), 1)
        self.assertEqual(memberships.keys(), [org])
        self.assertEqual(len(memberships[org]), 1)
        self.assertEqual(memberships[org][0].title, 'Director')

        with self.user('admin'):
            api.content.create(
                title='Leutenant',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury
            )

        memberships = IPerson(nick_fury).memberships()

        self.assertEqual(len(memberships), 1)
        self.assertEqual(memberships.keys(), [org])
        self.assertEqual(len(memberships[org]), 2)
        self.assertEqual(memberships[org][0].title, 'Director')
        self.assertEqual(memberships[org][1].title, 'Leutenant')



from collections import namedtuple
from datetime import date, datetime

from zope.interface import directlyProvides

from Products.ZCatalog.interfaces import ICatalogBrain

from plone.supermodel import loadString
from plone.namedfile.file import NamedImage
from plone.app.textfield.value import RichTextValue
from plone.uuid.interfaces import IUUID

from seantis.people.browser.renderer import Renderer
from seantis.people.utils import UUIDList, LinkList
from seantis.people import tests


class TestRenderer(tests.IntegrationTestCase):

    model = """<?xml version='1.0' encoding='utf8'?>
    <model xmlns="http://namespaces.plone.org/supermodel/schema">
        <schema>
            <field name="textline"
                   type="zope.schema.TextLine">
                <title>Textline</title>
            </field>
            <field name="text"
                   type="zope.schema.Text">
                <title>Text</title>
            </field>
            <field name="email"
                   type="seantis.plonetools.schemafields.Email">
                <title>Email</title>
            </field>
            <field name="richtext"
                    type="plone.app.textfield.RichText">
                <title>Richtext</title>
            </field>
            <field name="website"
                   type="seantis.plonetools.schemafields.Website">
                <title>Website</title>
            </field>
            <field name="blobimage"
                   type="plone.namedfile.field.NamedBlobImage">
                <title>Blobimage</title>
            </field>
            <field name="image"
                   type="plone.namedfile.field.NamedImage">
                <title>Image</title>
            </field>
            <field name="date"
                   type="zope.schema.Date">
                <title>Date</title>
            </field>
            <field name="datetime"
                   type="zope.schema.Datetime">
                <title>Datetime</title>
            </field>
            <field name="list" type="zope.schema.Set">
              <title>List</title>
              <value_type type="zope.schema.Choice">
                <values>
                  <element>One</element>
                  <element>Two</element>
                  <element>Three</element>
                </values>
              </value_type>
            </field>
        </schema>
    </model>"""

    @property
    def schema(self):
        return loadString(self.model).schema

    def render_value(self, field, value):
        renderer = Renderer(self.schema, place='list')
        context = namedtuple('MockContext', [field])(value)
        return renderer.render(context, field)

    def test_textline(self):
        self.assertEqual(self.render_value('textline', u'test'), u'test')

    def test_email(self):
        self.assertEqual(
            self.render_value('email', u'test@example.com'),
            u'<a href="mailto:test@example.com">test@example.com</a>'
        )
        self.assertEqual(self.render_value('email', None), u'')

    def test_website(self):
        self.assertEqual(
            self.render_value('website', u'http://example.com'),
            (
                u'<a href="http://example.com" '
                u'target="_blank">http://example.com</a>'
            )
        )
        self.assertEqual(self.render_value('website', None), u'')

    def test_text(self):
        self.assertEqual(
            self.render_value('text', 'One\nTwo'),
            u'One<br />Two'
        )
        self.assertEqual(
            self.render_value('text', 'One\r\nTwo'),
            u'One<br />Two'
        )

    def test_richtext(self):
        text = RichTextValue(raw="<b>This is important</b>")

        self.assertEqual(
            self.render_value('richtext', text),
            u'<b>This is important</b>'
        )

    def test_date(self):
        self.assertEqual(
            self.render_value('date', date(2014, 12, 15)),
            u'Dec 15, 2014'
        )

    def test_datetime(self):
        self.assertEqual(
            self.render_value('datetime', datetime(2014, 12, 15, 12, 0)),
            u'Dec 15, 2014 12:00 PM'
        )

    def test_image(self):

        renderer = Renderer(self.schema, place='list')

        class MockContext(object):
            def getURL(self):
                return u'http://nohost/mockbrain'

            def absolute_url(self):
                return u'http://nohost/mockobject'

        context = MockContext()
        context.image = None
        self.assertEqual(renderer.render(context, 'image'), '')

        # if the context does not provide ICatalogBrain, absolute_url is used
        context.image = NamedImage()
        self.assertEqual(
            renderer.render(context, 'image'),
            u'<img src="http://nohost/mockobject/@@images/image/tile" />'
        )

        # per default, the size on the detail view is different
        renderer = Renderer(self.schema, place='detail')
        self.assertEqual(
            renderer.render(context, 'image'),
            u'<img src="http://nohost/mockobject/@@images/image/mini" />'
        )

        # if the context provides ICatalogBrain, getURL is used
        directlyProvides(context, ICatalogBrain)
        self.assertEqual(
            renderer.render(context, 'image'),
            u'<img src="http://nohost/mockbrain/@@images/image/mini" />'
        )

    def test_uuid_list(self):
        self.assertEqual(self.render_value('uuids', UUIDList()), u'')

        folders = (self.new_temporary_folder(), self.new_temporary_folder())
        uuids = UUIDList(IUUID(folder) for folder in folders)

        self.assertIn(folders[0].title, self.render_value('uuids', uuids))
        self.assertIn(folders[1].title, self.render_value('uuids', uuids))
        self.assertIn(
            folders[0].absolute_url(), self.render_value('uuids', uuids)
        )
        self.assertIn(
            folders[1].absolute_url(), self.render_value('uuids', uuids)
        )

    def test_link_list(self):
        self.assertEqual(self.render_value('links', LinkList()), u'')

        links = LinkList([
            (u'Google', 'https://www.google.ch'),
            (u'Bing', 'http://lmgtfy.com/?q=bing')
        ])

        self.assertIn(
            u'<li><a href="https://www.google.ch">Google</a></li>',
            self.render_value('links', links)
        )

        self.assertIn(
            u'<li><a href="http://lmgtfy.com/?q=bing">Bing</a></li>',
            self.render_value('links', links)
        )



""" Provides a fast renderer for schema fields. Since this renderer is used
on lists and tables and therefore called many times in a request the following
options were not chosen:

z3c.form display mode rendering - too complicated, hard to customize and slow
zpt templates - quite a bit slower because of features we don't need

The snippets rendered here so simple that string.Template can be used which
is by far the fastest way of doing templates in python.

"""

import string

from zope.schema import (
    Bool,
    Decimal,
    FrozenSet,
    getFields,
    List,
    Set,
    Text,
    Tuple,
    Date,
    Datetime
)
from plone import api
from plone.namedfile.field import NamedBlobImage, NamedImage
from plone.app.textfield import RichText
from plone.app.uuid.utils import uuidToCatalogBrain
from Products.ZCatalog.interfaces import ICatalogBrain

from seantis.plonetools.schemafields import Email, Website
from seantis.plonetools import tools

from seantis.people import _
from seantis.people.supermodel import (
    get_list_render_options, get_detail_render_options
)
from seantis.people.utils import UUIDList, LinkList


class EmailFieldRenderer(object):

    template = string.Template(u'<a href="mailto:${mail}">${mail}</a>')

    def __call__(self, context, field, options):
        mail = getattr(context, field)
        if mail:
            return self.template.substitute(mail=mail)
        else:
            return u''


class WebsiteFieldRenderer(object):

    template = string.Template(u'<a href="${url}" target="_blank">${url}</a>')

    def __call__(self, context, field, options):
        url = getattr(context, field)
        if url:
            return self.template.substitute(url=url)
        else:
            return u''


class BoolRenderer(object):

    def __call__(self, context, field, options):
        value = getattr(context, field, False)

        if value is True:
            return _(u'Yes')
        else:
            return _(u'No')


class TextRenderer(object):

    def __call__(self, context, field, options):
        return '<br />'.join(getattr(context, field, u'').splitlines())


class DecimalRenderer(object):

    def __call__(self, context, field, options):
        value = getattr(context, field)

        if not value:
            return u''

        # a custom formatter, because the api of Zope's formatter is.. weird?
        formatter = context.REQUEST.locale.numbers.getFormatter('decimal')
        precision = options.by(field).get('precision', 'auto')

        if precision == 'auto':
            formatstring = u'{:,}'
        else:
            formatstring = u'{{:,.{}f}}'.format(precision)

        # this will fail miserably, if the group and the thousand separator
        # should ever be the same, but then again that's stupid
        output = formatstring.format(value)

        output = output.replace('.', 'd')
        output = output.replace(',', 'g')
        output = output.replace('d', formatter.symbols['decimal'])
        output = output.replace('g', formatter.symbols['group'])

        return output


class RichTextRenderer(object):

    def __call__(self, context, field, options):
        return getattr(context, field).output


class ListRenderer(object):

    def __call__(self, context, field, options):
        value = getattr(context, field, tuple())

        if not value:
            return u''

        render_type = options.by(field).get('list_type', 'comma-separated')

        if render_type == 'comma-separated':
            return u', '.join(getattr(context, field, tuple()))
        elif render_type == 'ul':
            return u'<ul class="dense"><li>{}</li></ul>'.format(
                u'</li><li>'.join(
                    value
                )
            )


class LinkListRenderer(object):

    template = string.Template(u'<li><a href="${url}">${title}</a></li>')

    def __call__(self, context, field, options):
        links = getattr(context, field, None)

        if not links:
            return u''

        return u'<ul class="dense">{}</ul>'.format(
            '\n'.join(
                self.template.substitute(url=l[1], title=l[0]) for l in links
            )
        )


class UUIDListRenderer(object):

    template = string.Template(u'<a href="${url}">${title}</a>')

    def __call__(self, context, field, options):
        uuids = getattr(context, field, None)

        if not uuids:
            return u''

        unicode_sortkey = tools.unicode_collate_sortkey()

        brains = (b for b in (uuidToCatalogBrain(uid) for uid in uuids) if b)
        items = sorted(
            ((b.getURL(), b.Title) for b in brains),
            key=lambda i: unicode_sortkey(i[1])
        )

        return ', '.join(
            self.template.substitute(url=url, title=title.decode('utf-8'))
            for url, title in items
        )


class ImageRenderer(object):

    template = string.Template(u'<img src="${url}" />')

    def __call__(self, context, field, options):
        img = getattr(context, field)
        if img:
            if ICatalogBrain.providedBy(context):
                baseurl = context.getURL()
            else:
                baseurl = context.absolute_url()

            size = options.by(field).get('image_size', 'thumb')
            url = '/'.join((baseurl, '@@images', field, size))

            return self.template.substitute(url=url)
        else:
            return u''


class DateRenderer(object):

    def __init__(self, long_format=False, time_only=False):
        self.long_format = long_format
        self.time_only = time_only

    def __call__(self, context, field, options):
        date = getattr(context, field)
        if date:
            return api.portal.get_localized_time(
                datetime=date,
                long_format=self.long_format,
                time_only=self.time_only
            )
        else:
            return u''


# This is not the best way to match objects to classes, but it sure is the
# fastest. Checking through isinstance would require going through a list.
# Since that can easily happen more than a thousand times in a request it's
# better to be fast than to be right.
renderers = {
    Email: EmailFieldRenderer(),
    Website: WebsiteFieldRenderer(),
    NamedBlobImage: ImageRenderer(),
    NamedImage: ImageRenderer(),
    Text: TextRenderer(),
    RichText: RichTextRenderer(),
    Decimal: DecimalRenderer(),
    Tuple: ListRenderer(),
    List: ListRenderer(),
    Set: ListRenderer(),
    FrozenSet: ListRenderer(),
    list: ListRenderer(),
    set: ListRenderer(),
    tuple: ListRenderer(),
    UUIDList: UUIDListRenderer(),
    LinkList: LinkListRenderer(),
    Bool: BoolRenderer(),
    Datetime: DateRenderer(long_format=True),
    Date: DateRenderer(long_format=False)
}


class Options(object):

    def __init__(self, field_options, default_options):
        self.field_options = field_options
        self.default_options = default_options

    def by(self, field):
        if field in self.field_options:
            return self.field_options[field]
        else:
            return self.default_options


class Renderer(object):

    def __init__(self, schema, place, redirects=None):
        assert place in ('list', 'detail')

        self.schema = schema
        self.fields = getFields(schema)
        self.redirects = redirects or {}

        if place == 'list':
            self.options = Options(get_list_render_options(schema), {
                'image_size': 'tile',
                'list_type': 'comma-separated',
                'precision': 'auto'
            })
        else:
            self.options = Options(get_detail_render_options(schema), {
                'image_size': 'mini',
                'list_type': 'comma-separated',
                'precision': 'auto'
            })

    def render(self, context, field):
        field = self.redirects.get(field, field)
        fieldtype = type(self.fields.get(field, getattr(context, field, None)))

        if fieldtype in renderers:
            return renderers[fieldtype](context, field, self.options)
        else:
            return getattr(context, field)



from plone import api

from seantis.people.supermodel import set_title_fields
from seantis.people import (
    behaviors,
    interfaces,
    tests
)


class TestBehavior(tests.IntegrationTestCase):

    def is_behavior_assignable(self, behavior, interface, factory):

        # use a throwaway type with the given behavior
        new_type = self.new_temporary_type(behaviors=[behavior])

        # create an object of said type
        with self.user('admin'):
            obj = api.content.create(
                id=new_type.id,
                type=new_type.id,
                container=self.new_temporary_folder()
            )

        # adapt the behavior's interface
        adapted = interface(obj)

        # ensure that an instance of the factory is the result
        self.assertIs(type(adapted), factory)
        self.assertIs(adapted.context, obj)

    def test_person_behavior_assignable(self):
        self.is_behavior_assignable(
            interfaces.IPerson.__identifier__,
            interfaces.IPerson,
            behaviors.Person
        )

    def test_person_title_no_influence_if_not_activated(self):
        blank_type = self.new_temporary_type()
        set_title_fields(blank_type.lookupSchema(), ['foo', 'bar'])

        with self.user('admin'):
            blank_obj = api.content.create(
                id='',
                type=blank_type.id,
                container=self.new_temporary_folder(),
                foo='stop',
                bar='hammertime!'
            )

        self.assertEqual(blank_obj.title, '')

    def test_person_title_behavior(self):
        new_type = self.new_temporary_type(
            behaviors=[
                interfaces.INameFromPerson.__identifier__
            ]
        )

        set_title_fields(new_type.lookupSchema(), ['foo', 'bar'])

        with self.user('admin'):
            obj = api.content.create(
                id='',
                type=new_type.id,
                container=self.new_temporary_folder(),
                foo='stop',
                bar='hammertime!'
            )

        self.assertEqual(obj.title, 'stop hammertime!')
        self.assertEqual(obj.id, 'stop-hammertime')

        obj.foo = 'start'
        behaviors.name_from_person.on_object_modified(obj)

        self.assertEqual(obj.title, 'start hammertime!')
        self.assertEqual(obj.id, 'stop-hammertime')


    def test_get_name_from_person(self):

        foobar_xml = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <field name="bar" type="zope.schema.TextLine">
                    <title>Bar</title>
                </field>
                <field name="foo" type="zope.schema.TextLine">
                    <title>Foo</title>
                </field>
            </schema>
        </model>"""

        person = self.new_temporary_type(
            model_source=foobar_xml
        )

        with self.user('admin'):
            obj = api.content.create(
                id='test',
                type=person.id,
                container=self.new_temporary_folder(),
                foo='mic check',
                bar='one two'
            )

        schema = person.lookupSchema()

        set_title_fields(schema, ['foo', 'bar'])
        self.assertEqual(
            behaviors.get_name_from_person(obj), u'mic check one two'
        )

        set_title_fields(schema, ['bar', 'foo'])
        self.assertEqual(
            behaviors.get_name_from_person(obj), u'one two mic check'
        )



from plone import api

from zope.security.management import newInteraction, endInteraction

from seantis.plonetools.testing import TestCase
from seantis.people.testing import (
    INTEGRATION_TESTING,
    FUNCTIONAL_TESTING
)


# to use with integration where security interactions need to be done manually
class IntegrationTestCase(TestCase):
    layer = INTEGRATION_TESTING

    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        # setup security
        newInteraction()

    def tearDown(self):
        endInteraction()
        super(IntegrationTestCase, self).tearDown()

    def get_import_form(self):
        with self.user('admin'):
            obj = api.content.create(
                id='people',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            return api.content.get_view('import', obj, self.request)


# to use with the browser which does its own security interactions
class FunctionalTestCase(TestCase):
    layer = FUNCTIONAL_TESTING



from contextlib import contextmanager

from zope.component import getGlobalSiteManager, getAdapters
from zope.interface import Interface
from zope.event import notify

from plone import api
from plone.uuid.interfaces import IUUID

from seantis.plonetools import tools

from seantis.people import catalog_id
from seantis.people import tests
from seantis.people.events import MembershipChangedEvent
from seantis.people.interfaces import IMembershipSource, IPerson

from seantis.people.utils import UUIDList


class TestAdapter(object):

    def __init__(self, context):
        self.context = context


class Membership(object):

    def __init__(self, person, role=None):
        self.person = person
        self.role = role


class TestMemberships(tests.IntegrationTestCase):

    def register_membership_source(self, adapter, interface, name):
        gsm = getGlobalSiteManager()
        gsm.registerAdapter(
            factory=adapter, required=(interface,), name=name,
            provided=IMembershipSource
        )

    def unregister_membership_source(self, adapter, interface, name):
        gsm = getGlobalSiteManager()
        gsm.unregisterAdapter(
            required=(interface,), name=name, provided=IMembershipSource
        )

    @contextmanager
    def custom_source(self, source):
        self.register_membership_source(source, Interface, 'custom-source')
        yield
        self.unregister_membership_source(source, Interface, 'custom-source')

    def get_test_person(self):
        person_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )

        with self.user('admin'):
            person = api.content.create(
                title='person',
                type=person_type.id,
                container=self.new_temporary_folder()
            )

        return person

    def test_membership_adapter_registration(self):

        class Source(TestAdapter):

            def memberships(self, person=None):
                pass

        with self.custom_source(Source):

            folder = self.new_temporary_folder()

            adapters = [
                name for name, source in
                getAdapters((folder, ), IMembershipSource)
            ]

            self.assertEqual(
                adapters, ['zodb-membership-source', 'custom-source']
            )

        adapters = [
            name for name, source in
            getAdapters((folder, ), IMembershipSource)
        ]
        self.assertEqual(adapters, ['zodb-membership-source'])

    def test_membership_organizations(self):

        person = self.get_test_person()
        organization = self.new_temporary_folder()

        class Source(TestAdapter):

            def memberships(self, person=None):
                return {
                    IUUID(organization): [
                        Membership(person)
                    ]
                }

        with self.custom_source(Source):
            self.assertEqual(IPerson(person).organizations(), [
                organization.title
            ])

            self.assertEqual(IPerson(person).organization_uuids(), UUIDList([
                IUUID(organization)
            ]))

    def test_current_role(self):
        person = self.get_test_person()
        organization = self.new_temporary_folder()

        class Source(TestAdapter):

            def memberships(self, person=None):
                return {
                    IUUID(organization): [
                        Membership(person, 'president'),
                        Membership(person, 'secretary')
                    ]
                }

        with self.custom_source(Source):
            memberships = IPerson(person).memberships()[IUUID(organization)]
            role = IPerson(person).current_role(memberships)

            self.assertEqual(role, 'president')

    def test_membership_changed(self):
        tools.add_attribute_to_metadata('organizations', catalog_id)

        organization = self.new_temporary_folder()

        person = self.get_test_person()
        memberships = [Membership(person, 'vice-president')]

        class Source(TestAdapter):

            def memberships(self, person=None):
                return {IUUID(organization): memberships}

        with self.custom_source(Source):
            catalog = api.portal.get_tool(catalog_id)
            rid = tools.get_brain_by_object(person, catalog_id).getRID()

            get_orgs = lambda: catalog.getMetadataForRID(rid)['organizations']

            # because the membership source was added after the person
            # was created, the organizations metadata column is empty
            self.assertEqual(get_orgs(), [])

            # after signalling that the membership has changed, the
            # organizations should be available
            notify(MembershipChangedEvent(person))
            self.assertEqual(get_orgs(), [organization.id])



from datetime import date, time, datetime

import isodate
import tablib

from zope.i18nmessageid.message import Message

from plone import api
from plone.app.textfield.value import RichTextValue
from plone.namedfile.file import NamedImage, NamedBlobImage

from seantis.plonetools import tools
from seantis.people import _
from seantis.people import catalog_id
from seantis.people.errors import ContentExportError


supported_formats = [
    'csv',
    'json',
    'xls',
    'xlsx'
]


def export_people(request, container, portal_type, fields,
                  review_state=None, include_inactive=True):

    people = get_people(container, portal_type, review_state, include_inactive)

    if not people:
        raise ContentExportError(_(u"No people to export"))

    records = (get_record(request, person, fields) for person in people)

    translate = tools.translator(request, 'seantis.people')
    dataset = tablib.Dataset(headers=[translate(f[1]) for f in fields])
    map(dataset.append, records)

    return dataset


def get_people(container, portal_type,
               review_state=None, include_inactive=True):

    catalog = api.portal.get_tool(catalog_id)
    path = '/'.join(container.getPhysicalPath())

    query = {
        'path': path,
        'portal_type': portal_type,
    }

    if review_state is not None:
        query['review_state'] = review_state

    if not include_inactive:
        query['is_active_person'] = True

    return [b.getObject() for b in catalog(query)]


def get_record(request, person, fields):
    record = []

    translate = tools.translator(request, 'seantis.people')

    for field in (get_field(person, f[0]) for f in fields):
        if isinstance(field, Message):
            record.append(translate(field))
        else:
            record.append(field)

    return record


def get_field(person, field):

    if not hasattr(person, field):
        raise ContentExportError(_(u"Field '${name}' does not exist", mapping={
            'name': field
        }))

    value = getattr(person, field)

    if value is None:
        return u''

    if isinstance(value, bool):
        return _(u'Yes') if value is True else _(u'No')

    if isinstance(value, basestring):
        return unicode(value)

    if field in person.membership_fields.keys():
        organizations = (api.content.get(UID=i).title for i in value.keys())
        return u', '.join(organizations)

    if isinstance(value, (list, tuple, set)):
        return u', '.join(value)

    if isinstance(value, dict):
        parts = []
        for key in sorted(value):
            parts.append(u'{}: {}'.format(key, value[key]))
        return u', '.join(parts)

    if isinstance(value, date):
        return unicode(isodate.date_isoformat(value))

    if isinstance(value, datetime):
        return unicode(isodate.datetime_isoformat(value))

    if isinstance(value, time):
        return unicode(isodate.time_isoformat(value))

    if isinstance(value, RichTextValue):
        return unicode(value.output)

    if isinstance(value, (NamedImage, NamedBlobImage)):
        base = person.absolute_url()

        if value.filename:
            return u'{}/@@download/{}/{}'.format(base, field, value.filename)
        else:
            return u'{}/@@download/{}'.format(base, field)

    return value



from datetime import datetime
from plone import api
from seantis.people import tests
from seantis.people.interfaces import IPerson
from seantis.people.supermodel import set_columns
from seantis.people.supermodel.indexing import on_type_modified


class TestCatalog(tests.IntegrationTestCase):

    def test_require_baseclass(self):
        without_baseclass = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
        )
        with_baseclass = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )

        on_type_modified(without_baseclass)
        on_type_modified(with_baseclass)

        # create one object with the correct baseclass, one without.
        # Both are going to be in portal_catalog, but the one with the
        # person baseclass will end up in the people catalog only.
        with self.user('admin'):
            container = self.new_temporary_folder()
            api.content.create(
                id='without-baseclass',
                type=without_baseclass.id,
                container=container
            )
            api.content.create(
                id='with-baseclass',
                type=with_baseclass.id,
                container=container
            )

        path = '/'.join(container.getPhysicalPath())
        portal_catalog = api.portal.get_tool('portal_catalog')
        people_catalog = api.portal.get_tool('seantis_people_catalog')

        self.assertEqual(
            [b.id for b in portal_catalog(path={'query': path, 'depth': 1})],
            ['without-baseclass', 'with-baseclass']
        )

        self.assertEqual(
            [b.id for b in people_catalog(path={'query': path, 'depth': 1})],
            ['with-baseclass']
        )

    def test_column_metadata(self):
        new_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )

        set_columns(new_type.lookupSchema(), [['foo']])
        on_type_modified(new_type)

        with self.user('admin'):
            api.content.create(
                id='123',
                type=new_type.id,
                container=self.new_temporary_folder(),
                foo='bar',
            )

        portal_catalog = api.portal.get_tool('portal_catalog')
        people_catalog = api.portal.get_tool('seantis_people_catalog')

        with self.assertRaises(AttributeError):
            portal_catalog(portal_type=new_type.id)[0].foo

        self.assertEqual(people_catalog(portal_type=new_type.id)[0].foo, 'bar')

    def test_default_indexes(self):
        new_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )

        set_columns(new_type.lookupSchema(), [['foo']])
        on_type_modified(new_type)

        with self.user('admin'):
            created = datetime.now()

            api.content.create(
                id='007',
                type=new_type.id,
                container=self.new_temporary_folder(),
                Subject='O O Seven',
                title='James Bond',
                description='Spy for his Majesty, the Queen',
                created=created
            )

        portal_catalog = api.portal.get_tool('portal_catalog')
        people_catalog = api.portal.get_tool('seantis_people_catalog')

        portal_brain = portal_catalog(portal_type=new_type.id)[0]
        people_brain = people_catalog(portal_type=new_type.id)[0]

        self.assertEqual(portal_brain.id, '007')
        self.assertEqual(portal_brain.Subject, 'O O Seven')
        self.assertEqual(portal_brain.Title, 'James Bond')
        self.assertEqual(portal_brain.created, created)
        self.assertEqual(
            portal_brain.Description, 'Spy for his Majesty, the Queen'
        )

        self.assertEqual(portal_brain.id, people_brain.id)
        self.assertEqual(portal_brain.Title, people_brain.Title)
        self.assertEqual(portal_brain.Subject, people_brain.Subject)
        self.assertEqual(portal_brain.created, people_brain.created)
        self.assertEqual(portal_brain.Description, people_brain.Description)

        self.assertEqual(len(portal_catalog(Subject='O O Seven')), 1)
        self.assertEqual(len(people_catalog(Subject='O O Seven')), 1)

        self.assertEqual(len(portal_catalog(SearchableText='James Bond')), 1)
        self.assertEqual(len(people_catalog(SearchableText='James Bond')), 1)

        # folder and person
        self.assertEqual(len(portal_catalog(created=created)), 2)

        # person only
        self.assertEqual(len(people_catalog(created=created)), 1)



import transaction
from copy import copy
from five import grok

from zope import schema
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm
from z3c.form import field
from z3c.form.browser.radio import RadioFieldWidget
from plone.directives import form
from plone.namedfile.field import NamedFile

from seantis.people import _
from seantis.people.errors import ContentImportError
from seantis.people.interfaces import IList
from seantis.people.browser import BaseForm
from seantis.people.content.import_content import (
    supported_formats,
    import_people
)


class IImportFormSchema(form.Schema):

    portal_type = schema.Choice(title=_(u"Type"), values=[])
    import_file = NamedFile(title=_(u"File"))


class ImportForm(BaseForm):

    grok.require('cmf.ManagePortal')
    grok.context(IList)
    grok.name('import')

    ignoreContext = True

    @property
    def label(self):
        return _(u"Import ${name}", mapping={
            'name': self.context.title
        })

    @property
    def fields(self):
        fields = field.Fields(IImportFormSchema)
        self.prepare_portal_type_field(fields)
        return fields

    @property
    def available_actions(self):
        yield dict(name='import', title=_(u'Import'), css_class='context')
        yield dict(name='cancel', title=_(u'Cancel'))

    def available_types_vocabulary(self):
        translate = lambda txt: self.translate(unicode(txt))
        return SimpleVocabulary(terms=[
            SimpleTerm(fti.id, fti.id, translate(fti.title)) for fti
            in self.context.available_types()
        ])

    def prepare_portal_type_field(self, fields):
        f = fields['portal_type']
        vocabulary = self.available_types_vocabulary()

        f.field = copy(f.field)
        f.field.vocabulary = vocabulary
        f.widgetFactory = RadioFieldWidget
        f.field.default = vocabulary._terms[0].value

    def get_format_from_filename(self, filename):
        if '.' in filename:
            return filename.split('.')[-1]
        else:
            return ''

    def handle_import(self):
        if not self.parameters:
            return

        try:
            portal_type = self.parameters['portal_type']
            import_file = self.parameters['import_file']

            format = self.get_format_from_filename(import_file.filename)

            if not format in supported_formats:
                raise ContentImportError(_(
                    u'${format} is not supported, supported are ${formats}',
                    mapping={
                        'format': format,
                        'formats': ', '.join(supported_formats)
                    }
                ))

            count = import_people(
                self.request,
                self.context,
                portal_type,
                format,
                import_file.data
            )

            self.status = _(
                u'Successfully imported ${count} records', mapping=dict(
                    count=count
                )
            )
        except ContentImportError, e:
            transaction.abort()
            self.raise_action_error(e.translate(self.request))
        else:
            self.request.response.redirect(self.context.absolute_url())



from plone.app.content.interfaces import INameFromTitle
from zope.interface import implements
from zope.component import adapts

from seantis.plonetools import tools
from seantis.people.supermodel import get_title_fields
from seantis.people.interfaces import INameFromPerson


def on_object_modified(obj, event=None):
    """ If an object with the INameFromPerson behavior activated is changed,
    the title needs to be changed. This does not happen automatically.

    """

    obj.setTitle(get_name_from_person(obj))


def get_name_from_person(obj):
    schema = tools.get_schema_from_portal_type(obj.portal_type)
    fields = get_title_fields(schema)

    title = u' '.join((getattr(obj, field, u'') or u'' for field in fields))

    return title.strip()


class NameFromPerson(object):
    """ Uses the fields defined as title by seantis.people.supermodel to
    generate a title for a new object.

    """
    implements(INameFromTitle)
    adapts(INameFromPerson)

    def __init__(self, context):
        pass

    def __new__(cls, context):
        title = get_name_from_person(context)
        instance = super(NameFromPerson, cls).__new__(cls)

        instance.title = title
        context.setTitle(title)

        return instance



import logging
log = logging.getLogger('seantis.people')

from zipfile import ZipFile

import csv
import os.path
import inspect

from StringIO import StringIO
from urllib2 import urlopen

from five import grok
from plone import api

from seantis.people import catalog_id
from seantis.people.browser import BaseView
from seantis.people.interfaces import IList
from seantis.plonetools import tools


class LoadTestRecordsView(BaseView):

    permission = 'cmf.ManagePortal'
    grok.require(permission)

    grok.context(IList)
    grok.name('load-test-records')

    typename = 'test-record'

    @property
    def count(self):
        return self.request.get('count', '500')

    @property
    def url(self):
        return 'http://www.briandunning.com/sample-data/{}.zip'.format(
            self.count
        )

    @property
    def csv(self):
        return '{}.csv'.format(self.count)

    def load_xml(self):
        path = os.path.dirname(
            os.path.abspath(
                inspect.getfile(inspect.currentframe())
            )
        )

        return open(os.path.join(path, 'load_test_records.xml'), 'r').read()

    def create_type(self):
        if tools.is_existing_portal_type(self.typename):
            return

        tools.add_new_dexterity_type(
            self.typename, model_source=self.load_xml(),
            title=u'Test Record',
            behaviors=[
                'seantis.people.interfaces.IPerson',
                'seantis.people.interfaces.INameFromPerson'
            ]
        )

    def import_sample_data(self):
        zipfile = ZipFile(StringIO(urlopen(self.url).read()))

        # get rid of windows newlines - there's probably a more efficient way
        content = '\n'.join(zipfile.open(self.csv).read().splitlines())
        reader = csv.reader(StringIO(content), delimiter=',')

        headers = None

        try:
            from collective.noindexing.patches import apply, unapply
            noindexing_available = True
            log.info("using collective.noindexing to speed the import up")
        except ImportError:
            noindexing_available = False
            log.info("install collective.noindexing for faster imports")

        if noindexing_available:
            apply()

        for rowix, row in enumerate(reader):
            if not headers:
                headers = [key.lower() for key in row]
            else:
                kwargs = dict(
                    (header, row[ix]) for ix, header in enumerate(headers)
                )

                api.content.create(
                    id='',
                    type=self.typename,
                    container=self.context,
                    **kwargs
                )

            if rowix % 100 == 0:
                log.info('imported 100 records')

        if noindexing_available:
            log.info('rebuilding the catalog')
            unapply()
            api.portal.get_tool(catalog_id).clearFindAndRebuild()

        return rowix

    def render(self):
        if not tools.in_debug_mode():
            return "zope must be run in debug mode for this to work"

        RESPONSE = self.request.RESPONSE
        RESPONSE.setHeader('Content-Type', 'text/plain;charset=utf-8')

        log.info('creating type')
        self.create_type()

        log.info('importing sample data')
        rows = self.import_sample_data()

        log.info('imported {} records'.format(rows))
        self.request.response.redirect(self.context.absolute_url())



# -*- coding: utf-8 -*-

from plone import api

from seantis.plonetools import tools

from seantis.people import catalog_id
from seantis.people import tests
from seantis.people.interfaces import IPerson
from seantis.people.supermodel import (
    set_selectable_fields, set_columns, get_compound_columns
)

from seantis.people.supermodel.indexing import (
    update_related_indexes,
    update_metadata
)


class TestIndex(tests.IntegrationTestCase):

    selectable_xml = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column selectable="true">
                        <people:item>first</people:item>
                    </people:column>
                    <people:column selectable="true">
                        <people:item>second</people:item>
                    </people:column>
                    <people:column selectable="true">
                        <people:item>third</people:item>
                    </people:column>

                </people:columns>
            </schema>
        </model>"""

    def test_selectable_fields_index(self):
        new_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            model_source=self.selectable_xml,
            klass='seantis.people.types.base.PersonBase'
        )
        schema = new_type.lookupSchema()

        catalog = api.portal.get_tool(catalog_id)

        ix = '{}_selectable_{}'

        self.assertIn(ix.format(new_type.id, 'first'), catalog.indexes())
        self.assertIn(ix.format(new_type.id, 'second'), catalog.indexes())
        self.assertIn(ix.format(new_type.id, 'third'), catalog.indexes())

        # nothing changes implicitly
        set_selectable_fields(schema, ['third'])

        self.assertIn(ix.format(new_type.id, 'first'), catalog.indexes())
        self.assertIn(ix.format(new_type.id, 'second'), catalog.indexes())
        self.assertIn(ix.format(new_type.id, 'third'), catalog.indexes())

        # the index must be rebuilt by hand (using the web-interface events
        # will take care of that)
        update_related_indexes(new_type)

        self.assertNotIn(ix.format(new_type.id, 'first'), catalog.indexes())
        self.assertNotIn(ix.format(new_type.id, 'second'), catalog.indexes())
        self.assertIn(ix.format(new_type.id, 'third'), catalog.indexes())

    def test_first_letter_index(self):
        self.login('admin')

        new_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            klass='seantis.people.types.base.PersonBase'
        )
        folder = self.new_temporary_folder()

        create = lambda title: api.content.create(
            title=title, type=new_type.id, container=folder
        )

        objects = [create(name) for name in (
            u'Andrew', u'Beat', u'cesar', u'Dexter', u'', u'ändu', u'Ödipus'
        )]

        catalog = api.portal.get_tool(catalog_id)
        index = catalog._catalog.getIndex('first_letter')

        values = [
            index.getEntryForObject(
                tools.get_brain_by_object(obj, catalog_id).getRID()
            ) for obj in objects
        ]

        self.assertEqual(sorted(values), [
            u'', u'A', u'B', u'C', u'D', u'Ä', u'Ö'
        ])

    def test_first_letter_index_update(self):
        foobar_xml = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:people="http://namespaces.plone.org/supermodel/people">
                <schema>
                    <people:order>
                        <people:item>foo</people:item>
                        <people:item>bar</people:item>
                    </people:order>
                </schema>
        </model>"""
        self.login('admin')

        foobar = self.new_temporary_type(
            behaviors=[IPerson.__identifier__],
            model_source=foobar_xml,
            klass='seantis.people.types.base.PersonBase'
        )

        obj = api.content.create(
            title='test',
            type=foobar.id,
            container=self.new_temporary_folder()
        )

        catalog = api.portal.get_tool(catalog_id)

        index = catalog._catalog.getIndex('first_letter')

        get_index_value = lambda: index.getEntryForObject(
            tools.get_brain_by_object(obj, catalog_id).getRID()
        )

        self.assertEqual(get_index_value(), u'T')

        obj.title = 'asdf'
        obj.reindexObject()
        self.assertEqual(get_index_value(), u'A')

        obj.bar = 'bar'
        obj.reindexObject()
        self.assertEqual(get_index_value(), u'B')

        obj.foo = u'ähm'
        obj.reindexObject()
        self.assertEqual(get_index_value(), u'Ä')

    def test_compound_columns_metadata(self):
        # Both key and value of a compound_column are added to the metadata
        # if the key alone is added.

        new_type = self.new_temporary_type(behaviors=[IPerson.__identifier__])

        compound_columns = get_compound_columns()
        columns = [[key] for key in compound_columns]
        set_columns(new_type.lookupSchema(), columns)

        catalog = api.portal.get_tool(catalog_id)

        # the sets are completely different, the compound keys/values are not
        # part of the catalog columns
        self.assertFalse(
            set(compound_columns.keys()) & set(catalog._catalog.names)
        )
        self.assertFalse(
            set(compound_columns.values()) & set(catalog._catalog.names)
        )

        update_metadata(new_type)

        # the compound keys/values are a subset of the catalog columns
        self.assertTrue(
            set(compound_columns.keys()) <= set(catalog._catalog.names)
        )
        self.assertTrue(
            set(compound_columns.values()) <= set(catalog._catalog.names)
        )



from collections import MutableSequence
from uuid import UUID


class UUIDList(MutableSequence):
    """ A class that acts exactly like a list, but only accepts UUIDS. In case
    of another 'typed' class or acute bordedom implement as a metaclass.

    """

    def __init__(self, data=None):
        self._list = list()

        if data is not None:
            map(self.append, data)

    def get_uuid(self, item):
        if not isinstance(item, UUID):
            return UUID(item).hex
        else:
            return item.hex

    def __eq__(self, other):
        if not hasattr(other, '_list'):
            return False
        return self._list == other._list

    def __len__(self):
        return len(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __delitem__(self, i):
        del self._list[i]

    def __setitem__(self, i, item):
        self._list[i] = self.get_uuid(item)

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return 'UUIDList({})'.format(repr(self._list))

    def insert(self, i, item):
        self._list.insert(i, self.get_uuid(item))

    def append(self, item):
        self.insert(len(self._list), self.get_uuid(item))


class LinkList(list):
    """ A class containing a list of tuples containing title, url. Used
    to render a <ul><li><a href="url">title</a></li></ul> list.

    """



# -*- coding: utf-8 -*-

from plone import api

from plone.uuid.interfaces import IUUID
from Products.CMFPlone.interfaces.constrains import IConstrainTypes, ENABLED

from seantis.people.interfaces import IPerson
from seantis.people.content import List
from seantis.people import tests
from seantis.people.content.list import ListConstrainTypes, LetterFilter
from seantis.people.utils import UUIDList

from pyquery import PyQuery as pq


class TestList(tests.IntegrationTestCase):

    def test_list_creation(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

        self.assertIs(type(lst.aq_base), List)

    def test_list_people_by_behavior(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            new_person_type = lambda: self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                klass='seantis.people.types.base.PersonBase'
            )

            person = new_person_type()
            new_person_type()  # intentionally unused

            # all types are possible before any are used
            self.assertEqual(len(lst.possible_types()), 2)
            self.assertEqual(len(lst.available_types()), 2)
            self.assertEqual(lst.used_type(), None)

            api.content.create(id='ned', type=person.id, container=lst)
            api.content.create(id='robb', type=person.id, container=lst)

            # only one type is used at this point
            self.assertEqual(len(lst.people()), 2)
            self.assertEqual(len(lst.possible_types()), 2)
            self.assertEqual(len(lst.available_types()), 1)
            self.assertEqual(lst.used_type(), person)

    def test_list_constrain(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )
            constrain = IConstrainTypes(lst)

            new_type = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                klass='seantis.people.types.base.PersonBase'
            )

            self.assertIs(type(constrain), ListConstrainTypes)
            self.assertEqual(constrain.getConstrainTypesMode(), ENABLED)
            self.assertEqual(constrain.allowedContentTypes(), [new_type])
            self.assertEqual(constrain.getLocallyAllowedTypes(), [new_type.id])
            self.assertEqual(
                constrain.getImmediatelyAddableTypes(), [new_type.id]
            )

    def test_add_unsupported_type(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            add_invalid_type = lambda: api.content.create(
                id='invalid', type=self.new_temporary_type(), container=lst
            )

        self.assertRaises(api.exc.InvalidParameterError, add_invalid_type)

    def test_list_letters(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                klass='seantis.people.types.base.PersonBase'
            )

            api.content.create(title='Jessie', type=person.id, container=lst)
            api.content.create(title='Walt', type=person.id, container=lst)

        self.assertEqual(sorted(lst.letters()), ['J', 'W'])

        filtered = lst.people(filter=LetterFilter('J', None))
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].id, 'jessie')

    def test_list_inactive(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                klass='seantis.people.types.base.PersonBase'
            )

            api.content.create(title='Abed', type=person.id, container=lst)
            api.content.create(
                title='Troy',
                type=person.id,
                container=lst,
                is_active_person=False
            )

        # admins see inactive people
        with self.user('admin'):
            self.assertEqual(len(lst.people()), 2)

        # normal users do not
        self.assertEqual(len(lst.people()), 1)

        # unless overriden by flag
        self.assertEqual(len(lst.people(include_inactive=True)), 2)

    def test_list_unrestricted(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                klass='seantis.people.types.base.PersonBase'
            )

            workflows = api.portal.get_tool('portal_workflow')
            workflows.setChainForPortalTypes(
                [person.id], 'simple_publication_workflow'
            )

            abed = api.content.create(
                title='Abed', type=person.id, container=lst
            )

            self.assertEqual(len(lst.people()), 1)

        # we are no longer admin so we don't see the private people
        self.assertEqual(len(lst.people(unrestricted_search=False)), 0)

        # unless we request it
        self.assertEqual(len(lst.people(unrestricted_search=True)), 1)

        # or we publish
        with self.user('admin'):
            api.content.transition(obj=abed, transition='publish')

        self.assertEqual(len(lst.people()), 1)

    def test_list_view_empty(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            view = lst.unrestrictedTraverse('@@view')

        self.assertIn('No people in the list.', view())

    def test_list_view_columns(self):

        model = """<?xml version='1.0' encoding='utf8'?>
        <model
            xmlns="http://namespaces.plone.org/supermodel/schema"
            xmlns:people="http://namespaces.plone.org/supermodel/people"
        >
            <schema>
                <people:columns>
                    <people:column>
                        <people:item>firstname</people:item>
                    </people:column>
                    <people:column>
                        <people:item>lastname</people:item>
                    </people:column>
                </people:columns>
                <field name="firstname" type="zope.schema.TextLine">
                    <title>First name</title>
                </field>
                <field name="lastname" type="zope.schema.TextLine">
                    <title>Last name</title>
                </field>
            </schema>
        </model>"""

        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder(),
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                model_source=model,
                klass='seantis.people.types.base.PersonBase'
            ).id

            api.content.create(
                id='miri', type=person, container=lst,
                firstname='Miriam', lastname='Linky'
            )
            api.content.create(
                id='zack', type=person, container=lst,
                firstname='Zack', lastname='Brown'
            )

            view = lst.unrestrictedTraverse('@@view')

        page = pq(view())

        self.assertEqual(len(page.find('thead th')), 2)

        self.assertEqual(len(page.find('tbody tr')), 2)
        self.assertEqual(len(page.find('tbody tr:first-child td')), 2)

        self.assertEqual(page.find('thead').text(), 'First name Last name')
        self.assertEqual(page.find('tbody').text(), 'Miriam Linky Zack Brown')

    def test_list_view_filter_multiple_values(self):
        model = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column selectable="true">
                        <people:item>countries</people:item>
                    </people:column>
                </people:columns>
            </schema>
        </model>"""

        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            group = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                model_source=model,
                klass='seantis.people.types.base.PersonBase'
            ).id

            countries = api.content.create(
                title='countries', type=group, container=lst, countries=[
                    'Germany', 'Switzerland'
                ]
            )

            view = lst.unrestrictedTraverse('@@view')

        columns = view.columns()

        view.request['filter-countries'] = 'Switzerland'

        self.assertEqual(len(view.people()), 1)
        self.assertEqual(view.column_values(columns[0]), countries.countries)
        self.assertEqual(view.selected_column_value(columns[0]), 'Switzerland')

        view.request['filter-countries'] = 'France'

        self.assertEqual(len(view.people()), 0)
        self.assertEqual(view.column_values(columns[0]), countries.countries)
        self.assertEqual(view.selected_column_value(columns[0]), 'France')

    def test_list_view_filter(self):

        model = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column selectable="true">
                        <people:item>country</people:item>
                    </people:column>
                </people:columns>
                <field name="country" type="zope.schema.TextLine">
                    <title>Country</title>
                </field>
            </schema>
        </model>"""

        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder(),
            )

            country = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                model_source=model,
                klass='seantis.people.types.base.PersonBase'
            ).id

            countries = ['Österreich', 'Germany', 'Switzerland']

            for c in countries:
                api.content.create(
                    title=c, type=country, container=lst, country=c
                )

            view = lst.unrestrictedTraverse('@@view')

        columns = view.columns()

        self.assertEqual(len(view.people()), 3)
        self.assertEqual(len(columns), 1)
        self.assertEqual(view.selected_column_value(columns[0]), '__all__')
        self.assertEqual(view.column_values(columns[0]), countries)

        view.request['filter-country'] = 'Switzerland'

        self.assertEqual(len(view.people()), 1)
        self.assertEqual(view.column_values(columns[0]), countries)
        self.assertEqual(view.selected_column_value(columns[0]), 'Switzerland')

        view.request['filter-country'] = 'Österreich'

        self.assertEqual(len(view.people()), 1)
        self.assertEqual(view.column_values(columns[0]), countries)
        self.assertEqual(view.selected_column_value(columns[0]), 'Österreich')

    def test_list_view_compound_column_organizations(self):

        model = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column selectable="true">
                        <people:item>organizations</people:item>
                    </people:column>
                </people:columns>
            </schema>
        </model>"""

        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder(),
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                model_source=model,
                klass='seantis.people.types.base.PersonBase'
            ).id

            organizations = [
                self.new_temporary_folder(),
                self.new_temporary_folder()
            ]

            # normally organizations are defined through memberships,
            # but it's easier here to just use class properties
            #
            # XXX -> this should be changed to really setup memberships, so
            # the testcode in PersonBase is not needed!
            api.content.create(
                title='test', type=person, container=lst,
                test_organizations=[o.title for o in organizations],
                test_organization_uuids=UUIDList(
                    [IUUID(o) for o in organizations]
                )
            )
            api.content.create(
                title='loner', type=person, container=lst,
                test_organizations=[], test_organization_uuids=UUIDList()
            )

            view = lst.unrestrictedTraverse('@@view')

        columns = view.columns()

        # no filter applied, both people shown
        self.assertEqual(len(view.people()), 2)

        # the select element contains the organizations
        self.assertEqual(
            view.column_values(columns[0]),
            sorted([o.title for o in organizations])
        )
        self.assertEqual(columns[0].filter_key, 'organizations')

        # the table contains the links to the organizations (from the uuids)
        page = pq(view())

        for org in organizations:
            link = page.find('.people-list a[href="{}"]'.format(
                org.absolute_url())
            )
            self.assertEqual(link.text(), org.title)

        # filtering the organizations happen without the uuids
        view.request['filter-organizations'] = organizations[0].title
        self.assertEqual(len(view.people()), 1)

    def test_list_view_schema(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            view = lst.unrestrictedTraverse('@@view')

        # don't call if there are no people
        self.assertRaises(AssertionError, lambda: view.schema)

        with self.user('admin'):
            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                klass='seantis.people.types.base.PersonBase'
            )
            api.content.create(id='test', type=person.id, container=lst)

        self.assertEqual(view.schema, person.lookupSchema())

    def test_list_view_filter_property(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            view = lst.unrestrictedTraverse('@@view')

        self.assertEqual(view.filter, None)

        view.filter_prefix = 'test-filter-'
        view.request['test-filter-key'] = 'value'

        self.assertEqual(view.filter.key, 'key')
        self.assertEqual(view.filter.value, 'value')
        self.assertEqual(view.filter.title, u'')



import logging
log = logging.getLogger('seantis.people')

from zope.i18nmessageid import MessageFactory
_ = MessageFactory('seantis.people')

# duplicated in profiles/default/toolset/toolset.xml
catalog_id = 'seantis_people_catalog'

import warnings
from exceptions import UnicodeWarning
warnings.filterwarnings('error', category=UnicodeWarning)



from uuid import uuid4
from seantis.people import tests, utils

class TestUtils(tests.IntegrationTestCase):

    def test_uuid_list(self):

        # works
        utils.UUIDList((uuid4(), uuid4()))

        # fails
        self.assertRaises(ValueError, lambda: utils.UUIDList(('test', )))



from plone import api
from zope.interface import implements

from seantis.plonetools import tools
from seantis.people.interfaces import IPerson
from seantis.people.behaviors.memberships import get_memberships
from seantis.people.utils import UUIDList


class Person(object):
    implements(IPerson)

    def __init__(self, context):
        self.context = context

    def memberships(self, org_filter=None):
        return get_memberships(person=self.context, org_filter=org_filter)

    def organizations(self, org_filter=None):
        """ Returns a list of organizations this person is active in. The list
        is sorted by uca collation. This is mainly used for the organizations
        metadata field of all person objects (to show in the list).

        """

        organizations = self.organization_uuids(org_filter)

        # organizations are defined in the general catalog, not the people
        # catalog, which only contains people, therefore use 'portal_catalog'
        catalog = api.portal.get_tool('portal_catalog')
        titles = (catalog(UID=uid)[0].Title for uid in organizations)

        return sorted(titles, key=tools.unicode_collate_sortkey())

    def organization_uuids(self, org_filter=None):
        """ Returns the organizations as uuid references which are be rendered
        as links in the list view.

        """
        organizations = UUIDList()

        for uuid, memberships in self.memberships().items():
            if callable(org_filter) and not org_filter(uuid, memberships):
                continue

            organizations.append(uuid)

        return organizations

    def current_role(self, memberships):
        """ Goes through the given memberships and returns the current role.
        The current role is the role of the first membership in the list.

        """

        if memberships:
            return memberships[0].role
        else:
            return u''



from lxml import etree

from plone.supermodel.parser import ISchemaMetadataHandler
from plone.supermodel.utils import ns
from zope.interface import implements

NAME_FROM_PERSON = u'seantis.people.name_from_person'
PERSON_COLUMNS = u'seantis.people.person_column'
PERSON_ORDER = u'seantis.people.order'
PERSON_SELECTABLE = u'seantis.people.selectable'
PERSON_DETAILS = u'seantis.people.details'
PERSON_LIST_RENDER_OPTIONS = u'seantis.people.list_render_options'
PERSON_DETAIL_RENDER_OPTIONS = u'seantis.people.detail_render_options'

# typo, but not really a problem, so just leave it, or face migration
PERSON_COLUMN_TITLE = u'senatis.people.column_title'

# Supermodel namespace and prefix
PEOPLE_NAMESPACE = 'http://namespaces.plone.org/supermodel/people'
PEOPLE_PREFIX = 'people'

missing = object()


def get_title_fields(schema):
    """ Get the fields used by INameFromPerson. """
    return schema.queryTaggedValue(NAME_FROM_PERSON, [])


def set_title_fields(schema, fields):
    """ Sets the fields used by INameFromPerson. """
    schema.setTaggedValue(NAME_FROM_PERSON, sorted(
        list(set(fields)), key=fields.index  # unique values, preserving order
    ))


def get_order(schema):
    """ Gets the fields by which the person list is ordered. """
    return schema.queryTaggedValue(PERSON_ORDER, [])


def set_order(schema, order):
    """ Sets the fields by which the person list is ordered. """
    schema.setTaggedValue(PERSON_ORDER, order)


def get_detail_fields(schema):
    """ Gets the fields shown in the detail view by position. E.g.

    {
        'left': ['firstname', 'lastname'],
        'right': ['image']
    }

    """
    return schema.queryTaggedValue(PERSON_DETAILS, {})


def set_detail_fields(schema, fields):
    """ Sets the fields shown in the detail view by position. """
    schema.setTaggedValue(PERSON_DETAILS, fields)


def get_columns(schema):
    return schema.queryTaggedValue(PERSON_COLUMNS, [])


def set_columns(schema, columns):
    for column in columns:
        assert not isinstance(column, basestring), """
            A list of lists is expected, not a list of strings.
        """
    schema.setTaggedValue(PERSON_COLUMNS, columns)


def get_selectable_fields(schema):
    return schema.queryTaggedValue(PERSON_SELECTABLE, [])


def set_selectable_fields(schema, fields):
    schema.setTaggedValue(PERSON_SELECTABLE, list(set(fields)))


def get_custom_column_titles(schema):
    return schema.queryTaggedValue(
        PERSON_COLUMN_TITLE, [None for col in get_columns(schema)]
    )


def set_custom_column_titles(schema, titles):
    schema.setTaggedValue(PERSON_COLUMN_TITLE, titles)


def set_list_render_options(schema, render_options):
    schema.setTaggedValue(PERSON_LIST_RENDER_OPTIONS, render_options)


def get_list_render_options(schema):
    return schema.queryTaggedValue(PERSON_LIST_RENDER_OPTIONS, {})


def set_detail_render_options(schema, render_options):
    schema.setTaggedValue(PERSON_DETAIL_RENDER_OPTIONS, render_options)


def get_detail_render_options(schema):
    return schema.queryTaggedValue(PERSON_DETAIL_RENDER_OPTIONS, {})


def render_options_to_dict(options):
    return dict(o.split('=') for o in options.split(';'))


def dict_to_render_options(options):
    return ';'.join(sorted('='.join((k, v)) for k, v in options.items()))


class NodeHandler(object):

    namespace = PEOPLE_NAMESPACE
    prefix = PEOPLE_PREFIX

    def __init__(self, tagname):
        self.tagname = tagname
        etree.register_namespace(self.prefix, self.namespace)

    def xpath(self, node, expression):
        return node.xpath(expression, namespaces={'people': self.namespace})

    def prefixed(self, name):
        return ns(name, self.namespace)

    def nodes(self, node, name):
        return node.findall('./{}'.format(self.prefixed(name)))

    def tags(self, node):
        return self.nodes(node, self.tagname)

    def items(self, node):
        return self.text(self.nodes(node, 'item'))

    def text(self, nodes):
        return [n.text.strip() for n in nodes]

    def render_options_from_items(self, parent):
        render_options = {}

        for item in self.nodes(parent, 'item'):
            if item.get('render-options'):
                render_options[item.text] = render_options_to_dict(
                    item.get('render-options')
                )

        return render_options


class ItemListHandler(NodeHandler):

    def __init__(self, tagname, getter, setter):
        super(ItemListHandler, self).__init__(tagname)
        self.getter = getter
        self.setter = setter

    def parse(self, schema_node, schema):
        expression = 'people:{}/people:item'.format(self.tagname)
        self.setter(schema, self.text(self.xpath(schema_node, expression)))

    def write(self, schema_node, schema):
        values = self.getter(schema)

        if not values:
            return

        element = etree.Element(self.prefixed(self.tagname))
        for value in values:
            item = etree.Element(self.prefixed('item'))
            item.text = value

            element.append(item)

        schema_node.append(element)


class DetailsHandler(NodeHandler):

    def parse(self, schema_node, schema):
        tags = self.tags(schema_node)

        if not tags:
            return

        detail_fields = {}
        render_options = {}

        for tag in tags:
            detail_fields[tag.get('position') or 'left'] = self.items(tag)
            render_options.update(self.render_options_from_items(tag))

        set_detail_fields(schema, detail_fields)
        set_detail_render_options(schema, render_options)

    def write(self, schema_node, schema):
        detail_fields = get_detail_fields(schema)
        render_options = get_detail_render_options(schema)

        if not detail_fields:
            return

        for position, fields in detail_fields.items():
            element = etree.Element(self.prefixed(self.tagname))
            element.set('position', position)

            for field in fields:
                item = etree.Element(self.prefixed('item'))
                item.text = field

                if field in render_options:
                    item.set('render-options', dict_to_render_options(
                        render_options[field]
                    ))

                element.append(item)

            schema_node.append(element)


class ColumnsHandler(NodeHandler):

    def parse(self, schema_node, schema):
        tags = self.tags(schema_node)

        if not tags:
            return

        columns, selectable_fields, titles = [], [], []
        render_options = {}

        for tag in tags:
            for column in self.nodes(tag, 'column'):
                column_items = self.items(column)

                if len(column_items) == 1 and column.get('selectable'):
                    selectable_fields.append(column_items[0])

                titles.append(column.get('title'))
                columns.append(column_items)

                render_options.update(self.render_options_from_items(column))

        set_columns(schema, columns)
        set_selectable_fields(schema, selectable_fields)
        set_custom_column_titles(schema, titles)
        set_list_render_options(schema, render_options)

    def write(self, schema_node, schema):
        columns = get_columns(schema)
        selectable_fields = get_selectable_fields(schema)
        titles = get_custom_column_titles(schema)
        render_options = get_list_render_options(schema)

        if not columns:
            return

        columns_el = etree.Element(self.prefixed(self.tagname))

        for ix, column in enumerate(columns):
            column_el = etree.Element(self.prefixed('column'))

            if len(column) == 1 and column[0] in selectable_fields:
                column_el.set('selectable', 'true')

            if titles[ix]:
                column_el.set('title', titles[ix])

            for field in column:
                item_el = etree.Element(self.prefixed('item'))
                item_el.text = field

                if field in render_options:
                    item_el.set('render-options', dict_to_render_options(
                        render_options[field]
                    ))

                column_el.append(item_el)

            columns_el.append(column_el)

        schema_node.append(columns_el)


class PeopleSchemaMetaHandler(object):

    implements(ISchemaMetadataHandler)

    namespace = PEOPLE_NAMESPACE
    prefix = PEOPLE_PREFIX

    handlers = [
        ItemListHandler('title', get_title_fields, set_title_fields),
        ItemListHandler('order', get_order, set_order),
        DetailsHandler('details'),
        ColumnsHandler('columns')
    ]

    def read(self, schema_node, schema):
        for handler in self.handlers:
            handler.parse(schema_node, schema)

    def write(self, schema_node, schema):
        for handler in self.handlers:
            handler.write(schema_node, schema)

    @staticmethod
    def register_utility():
        from zope.component import provideUtility

        from seantis.people.supermodel.schemahandler import (
            PeopleSchemaMetaHandler
        )
        provideUtility(
            component=PeopleSchemaMetaHandler(),
            name='seantis.people.people_schema_handler'
        )



from collections import namedtuple
from plone import api
from five import grok
from zope.interface import Interface

from seantis.plonetools import tools
from seantis.people.interfaces import IPerson
from seantis.people.browser.base import BaseView


class View(BaseView):
    """A number of macros for use with seantis.people"""

    grok.context(Interface)
    grok.require('zope2.View')
    grok.name('seantis-people-macros')

    template = grok.PageTemplateFile('templates/macros.pt')

    def __getitem__(self, key):
        return self.template._template.macros[key]

    def organizations(self, person, method):
        Organization = namedtuple('Organization', ['title', 'url', 'role'])
        catalog = api.portal.get_tool('portal_catalog')

        organizations = []

        for uuid, memberships in getattr(person, method).items():
            current_role = IPerson(person).current_role(memberships)

            brain = catalog(UID=uuid)[0]
            organizations.append(
                Organization(brain.Title, brain.getURL(), current_role)
            )

        sortkey = lambda org: tools.unicode_collate_sortkey()(org.title)
        return sorted(organizations, key=sortkey)



from collections import namedtuple

from five import grok
from zope.security import checkPermission as has_permission

from plone import api
from plone.dexterity.content import Container
from Products.CMFPlone.interfaces.constrains import IConstrainTypes, ENABLED

from seantis.people import catalog_id
from seantis.people.interfaces import IPerson, IList
from seantis.people.supermodel import get_selectable_field_ix
from seantis.plonetools import tools


ListFilter = namedtuple('ListFilter', ['key', 'value', 'title'])
LetterFilter = namedtuple('LetterFilter', ['value', 'title'])


class List(Container):

    def people(
        self,
        filter=None,
        batch_start=None,
        batch_size=None,
        include_inactive=False,
        unrestricted_search=False
    ):
        catalog = api.portal.get_tool(catalog_id)

        query = {}
        query['path'] = {
            'query': '/'.join(self.getPhysicalPath()), 'depth': 1
        }

        if not include_inactive:
            if not has_permission('cmf.ModifyPortalContent', self):
                query['is_active_person'] = True

        query['sort_on'] = 'sortable_title'
        query['sort_order'] = 'ascending'

        if batch_start is not None and batch_size is not None:
            query['b_start'] = batch_start
            query['b_size'] = batch_size

        if filter:
            used_type = self.used_type()

            if used_type:

                if isinstance(filter, ListFilter):
                    keyword_index = get_selectable_field_ix(
                        used_type.id, filter.key
                    )

                if isinstance(filter, LetterFilter):
                    keyword_index = 'first_letter'

                # the catalog cannot deal with unicode filter values
                query[keyword_index] = {'query': filter.value.encode('utf-8')}

        if unrestricted_search:
            return catalog.unrestrictedSearchResults(query)
        else:
            return catalog(query)

    def letters(self):
        catalog = api.portal.get_tool(catalog_id)
        index = catalog._catalog.getIndex('first_letter')

        letters = set()

        for person in self.people():
            letter = index.getEntryForObject(person.getRID())

            if letter:
                letters.add(letter)

        return sorted(letters, key=tools.unicode_collate_sortkey())

    def possible_types(self):
        return tools.get_type_info_by_behavior(IPerson.__identifier__)

    def available_types(self):
        used_type = self.used_type()

        if used_type:
            return [used_type]
        else:
            return self.possible_types()

    def used_type(self):
        catalog = api.portal.get_tool(catalog_id)
        path = {'query': '/'.join(self.getPhysicalPath()), 'depth': 1}

        for fti in self.possible_types():
            obj_of_type = catalog.unrestrictedSearchResults(
                path=path, portal_type=fti.id, sort_limit=1
            )
            if obj_of_type[:1]:
                return fti

        return None

    def reindex_organisations(self):
        for brain in self.people(unrestricted_search=True):
            brain.getObject().reindexObject(idxs=['membership_person'])


class ListConstrainTypes(grok.Adapter):
    grok.provides(IConstrainTypes)
    grok.context(IList)

    def getConstrainTypesMode(self):
        return ENABLED

    def allowedContentTypes(self):
        return self.context.available_types()

    def getLocallyAllowedTypes(self):
        return [t.id for t in self.allowedContentTypes()]

    def getImmediatelyAddableTypes(self):
        return self.getLocallyAllowedTypes()



from plone.directives import form
from seantis.people.types.base import PersonBase


class IPHZ(form.Schema):
    form.model("phz.xml")


class PHZ(PersonBase):

    @property
    def org_units(self):
        return [org for org in (self.org_unit_1, self.org_unit_2) if org]



from five import grok

from zope.component import getAdapters
from zope.interface import Interface
from zope.event import notify
from zope.lifecycleevent import ObjectModifiedEvent
from plone import api
from plone.uuid.interfaces import IUUID
from plone.indexer.decorator import indexer

from seantis.plonetools import tools
from seantis.people.interfaces import IMembershipSource, IMembership


@indexer(IMembership)
def membership_person(membership, **kw):
    """ Stores the person on each membership object for faster lookups. """
    if hasattr(membership.person, 'to_object'):
        person = membership.person.to_object
    else:
        person = membership.person

    return IUUID(person)


def on_membership_changed(event, *args, **kwargs):
    """ Listens to IMembershipChanged events relaying them to the Catalog
    for reindexing of the person whose membership changed.

    """
    notify(ObjectModifiedEvent(event.person))


def on_membership_content_item_changed(context, event):
    if hasattr(context.person, 'to_object'):
        notify(ObjectModifiedEvent(context.person.to_object))
    else:
        notify(ObjectModifiedEvent(context.person))


def get_memberships(person=None, org_filter=None):
    context = api.portal.get()

    organizations = {}
    for name, source in getAdapters((context, ), IMembershipSource):
        for organization, memberships in source.memberships(person).items():

            if callable(org_filter):
                if not org_filter(organization, memberships):
                    continue

            if organization not in organizations:
                organizations[organization] = []

            organizations[organization].extend(memberships)

    return organizations


class ZodbMembershipSource(grok.Adapter):

    grok.name('zodb-membership-source')
    grok.provides(IMembershipSource)
    grok.context(Interface)

    def memberships(self, person=None):

        query = {
            'object_provides': IMembership.__identifier__,
            'sort_on': 'getObjPositionInParent'
        }

        if person is not None:
            query['membership_person'] = IUUID(person)

        # memberships are stored outside the people's list, so the normal
        # catalog is used instead of the people's catalog
        memberships = api.portal.get_tool('portal_catalog')(**query)

        result = {}

        for membership in (m.getObject() for m in memberships):
            organization = IUUID(tools.get_parent(membership))
            result.setdefault(organization, []).append(membership)

        return result
